"""
Run Kmod
--------

Top-level script to analyse Kmod-results from the ``LHC`` and creating files for GUI and plotting
as well as returning beta-star and waist shift.
"""
from pathlib import Path

import numpy as np
import pandas as pd
import tfs
from generic_parser import EntryPointParameters, entrypoint

from omc3.kmod import analysis, helper
from omc3.kmod.constants import (BETA, ERR, EXT, FIT_PLOTS_NAME, INSTRUMENTS_FILE_NAME,
                                 LSA_FILE_NAME, RESULTS_FILE_NAME, SEQUENCES_PATH, STAR)
from omc3.utils import iotools, logging_tools

LOG = logging_tools.get_logger(__name__)

LSA_COLUMNS = ['NAME', f'{BETA}X', f'{ERR}{BETA}X', f'{BETA}Y', f'{ERR}{BETA}Y']


def kmod_params():
    parser = EntryPointParameters()
    parser.add_parameter(name='betastar_and_waist',
                         type=float,
                         required=True, nargs='+',
                         help='Estimated beta star of measurements and waist shift',)
    parser.add_parameter(name='working_directory',
                         type=Path,
                         required=True,
                         help='path to working directory with stored KMOD measurement files',)
    parser.add_parameter(name='beam',
                         type=int,
                         choices=[1, 2], required=True,
                         help='define beam used: 1 or 2',)
    parser.add_parameter(name='cminus', 
                         type=float,                         
                         help='C Minus',)
    parser.add_parameter(name='misalignment',
                         type=float,
                         help='misalignment of the modulated quadrupoles in m',)
    parser.add_parameter(name='errorK',
                         type=float,
                         help='error in K of the modulated quadrupoles, relative to gradient',)
    parser.add_parameter(name='errorL',
                         type=float,                         
                         help='error in length of the modulated quadrupoles, unit m',)
    parser.add_parameter(name='tune_uncertainty',
                         type=float,
                         default=2.5e-5,
                         help='tune measurement uncertainty')
    parser.add_parameter(name='instruments',
                         type=str,
                         default='MONITOR,SBEND,TKICKER,INSTRUMENT',
                         help='define instruments (use keywords from twiss) at which beta should '
                              'be calculated , separated by comma, e.g. MONITOR,RBEND,INSTRUMENT,TKICKER',)
    parser.add_parameter(name='simulation',
                         action='store_true',                         
                         help='flag for enabling simulation mode',)
    parser.add_parameter(name='log',
                         action='store_true',                         
                         help='flag for creating a log file')
    parser.add_parameter(name='no_autoclean',
                         action='store_true',
                         help='flag for manually cleaning data')
    parser.add_parameter(name='no_sig_digits',
                         action='store_true',                         
                         help='flag to not use significant digits')
    parser.add_parameter(name='no_plots',
                         action='store_true',                         
                         help='flag to not create any plots')
    parser.add_parameter(name='circuits',
                         type=str,
                         nargs=2,
                         help='circuit names of the modulated quadrupoles')
    parser.add_parameter(name='interaction_point',
                         type=str,
                         choices=['ip1', 'ip2', 'ip5', 'ip8', 'IP1', 'IP2', 'IP5', 'IP8'],
                         help='define interaction point')
    parser.add_parameter(name='measurement_dir',
                         type=Path,
                         help='give an optics measurement directory to include phase constraint in penalty function')
    parser.add_parameter(name='phase_weight',
                         type=float,
                         default=0.0,
                         help='weight in penalty function between phase and beta.'
                              'If weight=0 phase is not used as a constraint.')
    parser.add_parameter(name='model_dir',
                         type=Path,
                         help='twiss model that contains phase')
    parser.add_parameter(name="outputdir",
                         type=Path,
                         help="Path where outputfiles will be stored, defaults "
                                                "to the given working_directory")

    return parser


@entrypoint(kmod_params(), strict=True)
def analyse_kmod(opt):
    """
    Run Kmod analysis.

    Kmod Keyword Arguments:
        *--Required--*

        - **beam** *(int)*:

            define beam used: 1 or 2

            choices: ``[1, 2]``


        - **betastar_and_waist** *(float)*:

            Estimated beta star of measurements and waist shift


        - **working_directory** *(Path)*:

            path to working directory with stored KMOD measurement files


        *--Optional--*

        - **circuits** *(str)*:

            circuit names of the modulated quadrupoles


        - **cminus** *(float)*:

            C Minus


        - **errorK** *(float)*:

            error in K of the modulated quadrupoles, relative to gradient


        - **errorL** *(float)*:

            error in length of the modulated quadrupoles, unit m


        - **instruments** *(str)*:

            define instruments (use keywords from twiss) at which beta should be
            calculated , separated by comma, e.g. MONITOR,RBEND,INSTRUMENT,TKICKER

            default: ``MONITOR,SBEND,TKICKER,INSTRUMENT``


        - **interaction_point** *(str)*:

            define interaction point

            choices: ``['ip1', 'ip2', 'ip5', 'ip8', 'IP1', 'IP2', 'IP5', 'IP8']``


        - **log**:

            flag for creating a log file

            action: ``store_true``


        - **measurement_dir** *(Path)*:

            give an optics measurement directory to include phase constraint in
            penalty function


        - **misalignment** *(float)*:

            misalignment of the modulated quadrupoles in m


        - **model_dir** *(Path)*:

            twiss model that contains phase


        - **no_autoclean**:

            flag for manually cleaning data

            action: ``store_true``


        - **no_plots**:

            flag to not create any plots

            action: ``store_true``


        - **no_sig_digits**:

            flag to not use significant digits

            action: ``store_true``


        - **outputdir** *(Path)*:

            Path where outputfiles will be stored, defaults to the given
            working_directory


        - **phase_weight** *(float)*:

            weight in penalty function between phase and beta.If weight=0 phase is
            not used as a constraint.

            default: ``0.0``


        - **simulation**:

            flag for enabling simulation mode

            action: ``store_true``


        - **tune_uncertainty** *(float)*:

            tune measurement uncertainty

            default: ``2.5e-05``
    """
    LOG.info('Getting input parameter')
    if opt.interaction_point is None and opt.circuits is None:
        raise AttributeError('No IP or circuits specified, stopping analysis')
    if opt.interaction_point is not None and opt.circuits is not None:
        raise AttributeError('Both IP and circuits specified, choose only one, stopping analysis')
    if not 1 < len(opt.betastar_and_waist) < 5:
        raise AttributeError("Option betastar_and_waist has to consist of 2 to 4 floats")
    opt.betastar_and_waist = convert_betastar_and_waist(opt.betastar_and_waist)
    for error in ("cminus", "errorK", "errorL", "misalignment"):
        opt = check_default_error(opt, error)
    if opt.measurement_dir is None and opt.model_dir is None and opt.phase_weight:
        raise AttributeError("Cannot use phase advance without measurement or model")
    if opt.outputdir is None:
        opt.outputdir = opt.working_directory

    LOG.info(f"{'IP trim' if opt.interaction_point is not None else 'Individual magnets'} analysis")
    opt['magnets'] = MAGNETS_IP[opt.interaction_point.upper()] if opt.interaction_point is not None else [
        find_magnet(opt.beam, circuit) for circuit in opt.circuits]
    opt['label'] = f'{opt.interaction_point}B{opt.beam:d}' if opt.interaction_point is not None else f'{opt.magnets[0]}-{opt.magnets[1]}'
    opt['instruments'] = list(map(str.upper, opt.instruments.split(",")))

    output_dir = opt.outputdir / opt.label
    iotools.create_dirs(output_dir)

    LOG.info('Get inputfiles')
    magnet1_df, magnet2_df = helper.get_input_data(opt)
    opt, magnet1_df, magnet2_df, betastar_required = define_params(opt, magnet1_df, magnet2_df)

    LOG.info('Run simplex')
    magnet1_df, magnet2_df, results_df, instrument_beta_df = analysis.analyse(magnet1_df, magnet2_df, opt, betastar_required)

    LOG.info('Plot tunes and fit')
    if opt.no_plots:
        helper.plot_cleaned_data([magnet1_df, magnet2_df], output_dir / FIT_PLOTS_NAME, interactive_plot=False)

    LOG.info('Write magnet dataframes and results')
    for magnet_df in [magnet1_df, magnet2_df]:
        tfs.write(output_dir / f"{magnet_df.headers['QUADRUPOLE']}{EXT}", magnet_df)

    tfs.write(output_dir / f'{RESULTS_FILE_NAME}{EXT}', results_df)

    if opt.instruments_found:
        tfs.write(output_dir / f'{INSTRUMENTS_FILE_NAME}{EXT}', instrument_beta_df)

    create_lsa_results_file(betastar_required, opt.instruments_found, results_df, instrument_beta_df, output_dir)


def create_lsa_results_file(betastar_required, instruments_found, results_df, instrument_beta_df, output_dir):
    lsa_results_df = pd.DataFrame(columns=LSA_COLUMNS)
    if betastar_required:
        exporting_columns=['LABEL', f'{BETA}{STAR}X', f'{ERR}{BETA}{STAR}X', f'{BETA}{STAR}Y', f'{ERR}{BETA}{STAR}Y']
        lsa_results_df=results_df[exporting_columns].rename(columns=dict(zip(exporting_columns, LSA_COLUMNS)))
    if instruments_found:
        lsa_results_df = lsa_results_df.append(instrument_beta_df, sort=False, ignore_index=True)

    if not lsa_results_df.empty:
        tfs.write(output_dir / f'{LSA_FILE_NAME}{EXT}', lsa_results_df)


def convert_betastar_and_waist(bs):
    if len(bs) == 2:
        return dict(X=np.array([bs[0], bs[1]]), Y=np.array([bs[0], bs[1]]))
    if len(bs) == 3:
        return dict(X=np.array([bs[0], bs[2]]), Y=np.array([bs[1], bs[2]]))
    return dict(X=np.array([bs[0], bs[2]]), Y=np.array([bs[1], bs[3]]))


def check_default_error(options, error):
    if options[error] is None:
        options[error] = DEFAULTS_IP[error] if options.interaction_point is not None else DEFAULTS_CIRCUITS[error]
    return options


def find_magnet(beam, circuit):
    sequence = tfs.read(SEQUENCES_PATH / f"twiss_lhcb{beam:d}.dat")
    circuit = circuit.split('.')
    magnetname = sequence[sequence['NAME'].str.contains(r'MQ\w+\.{:s}{:s}{:s}\.\w+'.format(circuit[0][-1], circuit[1][0], circuit[1][1]))]['NAME'].to_numpy()[0]
    return magnetname


def define_params(options, magnet1_df, magnet2_df):
    LOG.debug(' adding additional parameters to header ')
    beta_star_required = False
    sequence = tfs.read(SEQUENCES_PATH / f"twiss_lhcb{options.beam:d}.dat", index='NAME')

    for magnet_df in [magnet1_df, magnet2_df]:
        magnet_df.headers['LENGTH'] = sequence.loc[magnet_df.headers['QUADRUPOLE'], 'L']
        magnet_df.headers['POLARITY'] = np.sign(sequence.loc[magnet_df.headers['QUADRUPOLE'], 'K1L'])

    magnet1_position_center = sequence.loc[magnet1_df.headers['QUADRUPOLE'], 'S'] - magnet1_df.headers['LENGTH']/2.
    magnet2_position_center = sequence.loc[magnet2_df.headers['QUADRUPOLE'], 'S'] - magnet2_df.headers['LENGTH']/2.

    ip_position = (magnet1_position_center + magnet2_position_center)/2.

    for magnet_df, magnet_position_center in zip([magnet1_df, magnet2_df], [magnet1_position_center, magnet2_position_center]):
        magnet_df.headers['LSTAR'] = np.abs(ip_position - magnet_position_center) - magnet_df.headers['LENGTH']/2.

    if magnet1_position_center < magnet2_position_center:
        left_magnet_df = magnet1_df
        right_magnet_df = magnet2_df
    elif magnet2_position_center < magnet1_position_center:
        left_magnet_df = magnet2_df
        right_magnet_df = magnet1_df

    between_magnets_df = sequence.reset_index().truncate(before=sequence.index.get_loc(left_magnet_df.headers['QUADRUPOLE']), after=sequence.index.get_loc(right_magnet_df.headers['QUADRUPOLE']), axis='index')

    if between_magnets_df.isin(['OMK']).any().loc['PARENT']:
        beta_star_required = True
    instruments = []
    for instrument in options.instruments:
        if between_magnets_df.isin([instrument]).any().loc['KEYWORD']:
            instruments.append(instrument)
            options[instrument] = dict(zip(between_magnets_df.loc[between_magnets_df['KEYWORD'] == instrument]['NAME'].to_numpy(),
                                           (between_magnets_df.loc[between_magnets_df['KEYWORD'] == instrument]['S'].to_numpy() - ip_position)))
    options.instruments_found = instruments
    return options, magnet1_df, magnet2_df, beta_star_required


DEFAULTS_IP = {
    "cminus": 1E-3,
    "misalignment": 0.006,
    "errorK": 0.001,
    "errorL": 0.001,
}

DEFAULTS_CIRCUITS = {
    "cminus": 1E-3,
    "misalignment": 0.001,
    "errorK": 0.001,
    "errorL": 0.001,
}

MAGNETS_IP = {
    "IP1": ['MQXA.1L1', 'MQXA.1R1'],
    "IP2": ['MQXA.1L2', 'MQXA.1R2'],
    "IP5": ['MQXA.1L5', 'MQXA.1R5'],
    "IP8": ['MQXA.1L8', 'MQXA.1R8']
}


if __name__ == '__main__':
    analyse_kmod()



"""
Constants
---------

Specific constants to be used in optics_measurements, to help with consistency.
"""
# File Names -------------------------------------------------------------------
EXT: str = ".tfs"
AMP_BETA_NAME: str = "beta_amplitude_"
BETA_NAME: str = "beta_phase_"
CHROM_BETA_NAME: str = "chrom_beta_"
PHASE_NAME: str = "phase_"
SPECIAL_PHASE_NAME: str = "special_phase_"
TOTAL_PHASE_NAME: str = "total_phase_"
DISPERSION_NAME: str = "dispersion_"
NORM_DISP_NAME: str = "normalised_dispersion_"
ORBIT_NAME: str = "orbit_"
KICK_NAME: str = "kick_"
IP_NAME: str = "interaction_point_"

# Column Names -----------------------------------------------------------------
# Pre- and Suffixe
ERR: str = "ERR"  # Error of the measurement
RMS: str = "RMS"  # Root-Mean-Square
RES: str = "RES"  # Rescaled measurement
DELTA: str = "DELTA"  # Delta between measurement and model (sometimes beating)
MDL: str = "MDL"  # Model
REAL: str = "REAL"
IMAG: str = "IMAG"

# Names
S: str = "S"
NAME: str = "NAME"
NAME2: str = f"{NAME}2"
TUNE: str = "Q"
NAT_TUNE: str = "NATQ"
PEAK2PEAK: str = "PK2PK"
ALPHA: str = "ALF"
BETA: str = "BET"
DPP: str = "DPP"
DPPAMP: str = "DPPAMP"
AMPLITUDE: str = "AMP"
PHASE: str = "PHASE"
PHASE_ADV: str = "MU"
F1001: str = "F1001"
F1010: str = "F1010"

SECONDARY_AMPLITUDE_X: str = "AMP01_X"  # amplitude of secondary line in horizontal spectrum
SECONDARY_AMPLITUDE_Y: str = "AMP10_Y"  # amplitude of secondary line in vertical spectrum
SECONDARY_FREQUENCY_X: str = "PHASE01_X"  # frequency of secondary line in horizontal spectrum
SECONDARY_FREQUENCY_Y: str = "PHASE10_Y"  # frequency of secondary line in vertical spectrum

TIME: str = "TIME"
ACTION: str = "2J"
SQRT_ACTION: str = "sqrt2J"


# Headers ----------------------------------------------------------------------
RESCALE_FACTOR: str = "RescalingFactor"



r"""
Knob Extractor
--------------

Entrypoint to extract knobs from ``NXCALS`` at a given time, and eventually write them out to a file.
This script can also be used to display information about the StateTracker State.
The data is fetched from ``NXCALS`` through ``pytimber`` using the **StateTracker** fields.

.. note::
    Please note that access to the GPN is required to use this functionality.

**Arguments:**

*--Optional--*

- **knob_definitions** *(PathOrStrOrDataFrame)*:

    User defined path to the knob-definitions, or (via python) a dataframe
    containing the knob definitions with the columns 'madx', 'lsa' and
    'scaling'.


- **knobs** *(str)*:

    A list of knob names or categories to extract. Available categories
    are: sep, xing, chroma, ip_offset, disp, mo.

    default: ``['sep', 'xing', 'chroma', 'ip_offset', 'disp', 'mo']``


- **output** *(PathOrStr)*:

    Specify user-defined output path. This should probably be
    `model_dir/knobs.madx`

    default: ``knobs.madx``


- **state**:

    Prints the state of the StateTracker. Does not extract anything else.

    action: ``store_true``


- **time** *(str)*:

    At what time to extract the knobs. Accepts ISO-format (YYYY-MM-
    DDThh:mm:ss), timestamp or 'now'. The default timezone for the ISO-
    format is local time, but you can force e.g. UTC by adding +00:00.

    default: ``now``


- **timedelta** *(str)*:

    Add this timedelta to the given time. The format of timedelta is
    '((\d+)(\w))+' with the second token being one of s(seconds),
    m(minutes), h(hours), d(days), w(weeks), M(months) e.g 7m = 7 minutes,
    1d = 1day, 7m30s = 7 min 30 secs. A prefix '_' specifies a negative
    timedelta. This allows for easily getting the setting e.g. 2h ago:
    '_2h' while setting the `time` argument to 'now' (default).


"""
import argparse
import logging
import math
import re
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

import pandas as pd
from dateutil.relativedelta import relativedelta

import tfs
from generic_parser import EntryPointParameters, entrypoint
from omc3.utils.iotools import PathOrStr, PathOrStrOrDataFrame
from omc3.utils.logging_tools import get_logger
from omc3.utils.mock import cern_network_import

pytimber = cern_network_import("pytimber")

LOGGER = get_logger(__name__)

AFS_ACC_MODELS_LHC = Path("/afs/cern.ch/eng/acc-models/lhc/current")
ACC_MODELS_LHC = Path("acc-models-lhc")
KNOBS_FILE_ACC_MODELS = ACC_MODELS_LHC / "operation" / "knobs.txt"
KNOBS_FILE_AFS = AFS_ACC_MODELS_LHC / "operation" / "knobs.txt"

MINUS_CHARS: Tuple[str, ...] = ("_", "-")


class Col:
    """ DataFrame Columns used in this script. """
    madx: str = "madx"
    lsa: str = "lsa"
    scaling: str = "scaling"
    value: str = "value"


class Head:
    """ TFS Headers used in this script."""
    time: str = "EXTRACTION_TIME"


KNOB_CATEGORIES: Dict[str, List[str]] = {
    "sep": [
        "LHCBEAM:IP1-SEP-H-MM",
        "LHCBEAM:IP1-SEP-V-MM",
        "LHCBEAM:IP5-SEP-H-MM",
        "LHCBEAM:IP5-SEP-V-MM",
    ],
    "xing": [
        "LHCBEAM:IP1-XING-V-MURAD",
        "LHCBEAM:IP1-XING-H-MURAD",
        "LHCBEAM:IP5-XING-V-MURAD",
        "LHCBEAM:IP5-XING-H-MURAD",
    ],
    "chroma": [
        "LHCBEAM1:QPH",
        "LHCBEAM1:QPV",
        "LHCBEAM2:QPH",
        "LHCBEAM2:QPV",
    ],
    "ip_offset": [
        "LHCBEAM:IP1-OFFSET-V-MM",
        "LHCBEAM:IP2-OFFSET-V-MM",
        "LHCBEAM:IP5-OFFSET-H-MM",
        "LHCBEAM:IP8-OFFSET-H-MM",
    ],
    "disp": [
        "LHCBEAM:IP1-SDISP-CORR-SEP",
        "LHCBEAM:IP1-SDISP-CORR-XING",
        "LHCBEAM:IP5-SDISP-CORR-SEP",
        "LHCBEAM:IP5-SDISP-CORR-XING",
    ],
    "mo": [
        "LHCBEAM1:LANDAU_DAMPING",
        "LHCBEAM2:LANDAU_DAMPING"
    ]
}

USAGE_EXAMPLES = """Usage Examples:

python -m omc3.knob_extractor --knobs disp chroma --time 2022-05-04T14:00     
    extracts the chromaticity and dispersion knobs at 14h on May 4th 2022

python -m omc3.knob_extractor --knobs disp chroma --time now _2h 
    extracts the chromaticity and dispersion knobs as of 2 hours ago

python -m omc3.knob_extractor --state
    prints the current StateTracker/State metadata

python -m omc3.knob_extractor --knobs disp sep xing chroma ip_offset mo --time now
    extracts the current settings for all the knobs
"""


def get_params():
    return EntryPointParameters(
        knobs=dict(
            type=str,
            nargs='*',
            help=(
                "A list of knob names or categories to extract. "
                f"Available categories are: {', '.join(KNOB_CATEGORIES.keys())}."
            ),
            default=list(KNOB_CATEGORIES.keys()),
        ),
        time=dict(
            type=str,
            help=(
                "At what time to extract the knobs. "
                "Accepts ISO-format (YYYY-MM-DDThh:mm:ss), timestamp or 'now'. "
                "The default timezone for the ISO-format is local time, "
                "but you can force e.g. UTC by adding +00:00."
            ),
            default="now",
        ),
        timedelta=dict(
            type=str,
            help=(
                "Add this timedelta to the given time. "
                "The format of timedelta is '((\\d+)(\\w))+' "
                "with the second token being one of "
                "s(seconds), m(minutes), h(hours), d(days), w(weeks), M(months) "
                "e.g 7m = 7 minutes, 1d = 1day, 7m30s = 7 min 30 secs. "
                "A prefix '_' specifies a negative timedelta. "
                "This allows for easily getting the setting "
                "e.g. 2h ago: '_2h' while setting the `time` argument to 'now' (default)."
            ),
        ),
        state=dict(
            action='store_true',
            help=(
                "Prints the state of the StateTracker. "
                "Does not extract anything else."
            ),
        ),
        output=dict(
            type=PathOrStr,
            help=(
                "Specify user-defined output path. "
                "This should probably be `model_dir/knobs.madx`"
            ),
        ),
        knob_definitions=dict(
            type=PathOrStrOrDataFrame,
            help=(
                "User defined path to the knob-definitions, "
                "or (via python) a dataframe containing the knob definitions "
                "with the columns 'madx', 'lsa' and 'scaling'."
            ),
        ),
    )


@entrypoint(
    get_params(), strict=True,
    argument_parser_args=dict(
        epilog=USAGE_EXAMPLES,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        prog="Knob Extraction Tool."
    )
)
def main(opt) -> Optional[tfs.TfsDataFrame]:
    """ Main knob extracting function. """
    ldb = pytimber.LoggingDB(source="nxcals", loglevel=logging.ERROR)
    time = _parse_time(opt.time, opt.timedelta)

    if opt.state:
        # only print the state of the StateTracker - the MetaState!
        # I still don't know what this does,
        # because I only get back that the variable does not exist. (jdilly)
        state = ldb.get("LhcStateTracker:State", time)  # do first to have output together
        LOGGER.info("---- STATE ------------------------------------")
        LOGGER.info(state)
        return None
    
    knobs_dict = _parse_knobs_defintions(opt.knob_definitions)
    knobs_extract = _extract_and_gather(ldb, knobs_dict, opt.knobs, time)
    if opt.output:
        _write_knobsfile(opt.output, knobs_extract)
    return knobs_extract


def extract(ldb, knobs: Sequence[str], time: datetime) -> Dict[str, float]:
    """
    Standalone function to gather data from  the StateTracker.
    Extracts data via pytimber's LoggingDB for the knobs given
    (either by name or by category) in knobs.

    Args:
        ldb (pytimber.LoggingDB): The pytimber database.
        knobs (Sequence[str]): Knob Categories or Knob-Names to extract.
        time (datetime): The time, when to extract.

    Returns:
        Dict[str, float]: Contains all the extracted knobs.
        When extraction was not possible, the value is None.

    """
    LOGGER.info(f"---- EXTRACTING KNOBS @ {time} ----")
    knobs_extracted = {}

    for category in knobs:
        for knob in KNOB_CATEGORIES.get(category, [category]):
            knobkey = f"LhcStateTracker:{knob}:target"
            knobs_extracted[knob] = None  # to log that this was tried to be extracted.

            knobvalue = ldb.get(knobkey, time.timestamp())  # use timestamp to preserve timezone info
            if knobkey not in knobvalue:
                LOGGER.warning(f"No value for {knob} found")
                continue

            timestamps, values = knobvalue[knobkey]
            if len(values) == 0:
                LOGGER.debug(f"No value for {knob} found")
                continue

            value = values[-1]
            if not math.isfinite(value):
                LOGGER.debug(f"Value for {knob} is not a number or infinite")
                continue

            LOGGER.info(f"Knob value for {knob} extracted: {value} (unscaled)")
            knobs_extracted[knob] = value

    return knobs_extracted


def check_for_undefined_knobs(knobs_definitions: pd.DataFrame, knob_categories: Sequence[str]):
    """ Check that all knobs are actually defined in the knobs-definitions.


    Args:
        knobs_definitions (pd.DataFrame): A mapping of all knob-names to KnobEntries.
        knob_categories (Sequence[str]): Knob Categories or Knob-Names to extract.

    Raises:
        KeyError: If one or more of the knobs don't have a definition.

    """
    knob_names = [knob for category in knob_categories for knob in KNOB_CATEGORIES.get(category, [category])]
    undefined_knobs = [knob for knob in knob_names if knob not in knobs_definitions.index]
    if undefined_knobs:
        raise KeyError(
            "The following knob(s) could not be found "
            f"in the knob-definitions: '{', '.join(undefined_knobs)}'"
        )


def _extract_and_gather(ldb, knobs_definitions: pd.DataFrame,
                        knob_categories: Sequence[str],
                        time: datetime) -> tfs.TfsDataFrame:
    """
    Main function to gather data from the StateTracker and the knob-definitions.
    All given knobs (either in categories or as knob names) to be extracted
    are checked for being present in the ``knob_definitions``.
    A TfsDataFrame is returned, containing the knob-definitions of the
    requested knobs and the extracted value (or NAN if not successful).

    Args:
        ldb (pytimber.LoggingDB): The pytimber database.
        knobs_definitions (pd.DataFrame): A mapping of all knob-names to KnobEntries.
        knob_categories (Sequence[str]): Knob Categories or Knob-Names to extract.
        time (datetime): The time, when to extract.

    Returns:
        tfs.TfsDataframe: Contains all the extracted knobs, in columns containing
        their madx-name, lsa-name, scaling and extracted value.
        When extraction was not possible, the value of the respective entry is NAN.

    """
    check_for_undefined_knobs(knobs_definitions, knob_categories)
    extracted_knobs = extract(ldb, knobs=knob_categories, time=time)

    knob_names = list(extracted_knobs.keys())
    knobs = tfs.TfsDataFrame(index=knob_names,
                             columns=[Col.lsa, Col.madx, Col.scaling, Col.value],
                             headers={Head.time: time})
    knobs[[Col.lsa, Col.madx, Col.scaling]] = knobs_definitions.loc[knob_names, :]
    knobs[Col.value] = pd.Series(extracted_knobs)
    return knobs


def _write_knobsfile(output: Union[Path, str], collected_knobs: tfs.TfsDataFrame):
    """ Takes the collected knobs and writes them out into a text-file. """
    collected_knobs = collected_knobs.copy()  # to not modify the df

    # Sort the knobs by category
    category_knobs = {}
    for category, category_names in KNOB_CATEGORIES.items():
        names = [name for name in collected_knobs.index if name in category_names]
        if not names:
            continue

        category_knobs[category] = collected_knobs.loc[names, :]
        collected_knobs = collected_knobs.drop(index=names)

    if len(collected_knobs):  # leftover knobs without category
        category_knobs["Other Knobs"] = collected_knobs

    # Write them out
    with open(output, "w") as outfile:
        outfile.write(f"!! --- knobs extracted by knob_extractor\n")
        outfile.write(f"!! --- extracted knobs for time {collected_knobs.headers[Head.time]}\n\n")
        for category, knobs_df in category_knobs.items():
            outfile.write(f"!! --- {category:10} --------------------\n")
            for knob, knob_entry in knobs_df.iterrows():
                outfile.write(f"{get_madx_command(knob_entry)}\n")
            outfile.write("\n")
        outfile.write("\n")


# Knobs Definitions ------------------------------------------------------------

def _get_knobs_def_file(user_defined: Optional[Union[Path, str]] = None) -> Path:
    """ Check which knobs-definition file is appropriate to take. """
    if user_defined is not None:
        LOGGER.info(f"Using user defined knobs.txt: '{user_defined}")
        return Path(user_defined)

    if KNOBS_FILE_ACC_MODELS.is_file():
        LOGGER.info(f"Using model folder's knobs.txt: '{KNOBS_FILE_ACC_MODELS}")
        return KNOBS_FILE_ACC_MODELS

    if KNOBS_FILE_AFS.is_file():
        # if all fails, fall back to lhc acc-models
        LOGGER.info(f"Using fallback knobs.txt: '{KNOBS_FILE_AFS}'")
        return KNOBS_FILE_AFS

    raise FileNotFoundError("None of the knobs-definition files are available.")


def load_knobs_definitions(file_path: Union[Path, str]) -> pd.DataFrame:
    """ Load the knobs-definition file and convert into a DataFrame.
    Each line in this file should consist of at least three comma separated
    entries in the following order: madx-name, lsa-name, scaling factor.
    Other columns are ignored.
    Alternatively, a TFS-file is also allowed, but needs to have the suffix ``.tfs``.

    Args:
        file_path (Path): Path to the knobs definition file.

    Returns:
        Dataframe with LSA names (but with colon instead of /) as
        keys and KnobEntries (without values) as value.
    """
    if Path(file_path).suffix == ".tfs":
        # just in case someone wants to give tfs files (hidden feature)
        df = tfs.read_tfs(file_path)
    else:
        # parse csv file (the official way)
        converters = {Col.madx: str.strip, Col.lsa: str.strip}  # strip whitespaces
        dtypes = {Col.scaling: float}
        names = (Col.madx, Col.lsa, Col.scaling)
        df = pd.read_csv(file_path,
                         comment="#",
                         usecols=list(range(len(names))),  # only read the first columns
                         names=names,
                         dtype=dtypes,
                         converters=converters)
    return _to_knobs_dataframe(df)


def _to_knobs_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """ Adapts a DataFrame to the conventions used here:
    StateTracker variable name as index, all columns lower-case.

    Args:
        df (pd.DataFrame): DataFrame containing at least the columns
                           'lsa', 'madx', 'scaling' (upper or lowercase)

    Returns:
        Dataframe with LSA names (but with colon instead of /) as
        keys and 'lsa', 'madx', 'scaling' and (empty) 'value' columns.

    """
    df.columns = df.columns.astype(str).str.lower()
    df = df[[Col.lsa, Col.madx, Col.scaling]].set_index(Col.lsa, drop=False)
    df.index = df.index.map(lsa2name)
    return df


def _parse_knobs_defintions(knobs_def_input: Optional[Union[Path, str, pd.DataFrame]]) -> pd.DataFrame:
    """ Parse the given knob-definitions either from a csv-file or from a DataFrame. """
    if isinstance(knobs_def_input, pd.DataFrame):
        return _to_knobs_dataframe(knobs_def_input)

    # input points to a file or is None
    knobs_def_file = _get_knobs_def_file(knobs_def_input)
    return load_knobs_definitions(knobs_def_file)


def get_madx_command(knob_data: pd.Series) -> str:
    if Col.value not in knob_data.index:
        raise KeyError("Value entry not found in extracted knob_data. "
                       "Something went wrong as it should at least be NaN.")
    if knob_data[Col.value] is None or pd.isna(knob_data[Col.value]):
        return f"! {knob_data[Col.madx]} : No Value extracted"
    return f"{knob_data[Col.madx]} := {knob_data[Col.value] * knob_data[Col.scaling]};"


# Time Tools -------------------------------------------------------------------

def _parse_time(time: str, timedelta: str = None) -> datetime:
    """ Parse time from given time-input. """
    t = _parse_time_from_str(time)
    if timedelta:
        t = _add_time_delta(t, timedelta)
    return t


def _parse_time_from_str(time_str: str) -> datetime:
    """ Parse time from given string. """
    # Now? ---
    if time_str.lower() == "now":
        return datetime.now()

    # ISOFormat? ---
    try:
        return datetime.fromisoformat(time_str)
    except (TypeError, ValueError):
        LOGGER.debug("Could not parse time string as ISO format")
        pass

    # Timestamp? ---
    try:
        return datetime.fromtimestamp(int(time_str))
    except (TypeError, ValueError):
        LOGGER.debug("Could not parse time string as a timestamp")
        pass

    raise ValueError(f"Couldn't read datetime '{time_str}'")


def _add_time_delta(time: datetime, delta_str: str) -> datetime:
    """ Parse delta-string and add time-delta to time. """
    sign = -1 if delta_str[0] in MINUS_CHARS else 1
    all_deltas = re.findall(r"(\d+)(\w)", delta_str)  # tuples (value, timeunit-char)

    # mapping char to the time-unit as accepted by relativedelta,
    # following ISO-8601 for time durations
    char2unit = dict(
        s='seconds', m='minutes', h='hours',
        d='days', w='weeks', M='months', Y="years",
    )

    # add all deltas, which are tuples of (value, timeunit-char)
    time_parts = {char2unit[delta[1]]: sign * int(delta[0]) for delta in all_deltas}
    time = time + relativedelta(**time_parts)

    return time


# Other tools ------------------------------------------------------------------

def lsa2name(lsa_name: str) -> str:
    """LSA name -> Variable in Timber/StateTracker conversion."""
    return lsa_name.replace("/", ":")


def name2lsa(name: str) -> str:
    """Variable in Timber/StateTracker -> LSA name conversion."""
    return name.replace(":", "/")


if __name__ == "__main__":
    main()



"""
Response Creator
----------------

Provides a response generation wrapper.
The response matrices can be either created by :mod:`omc3.correction.response_madx`
or analytically via :mod:`omc3.correction.response_twiss`.

Input arguments are split into response creation arguments and accelerator arguments.
The former are listed below, the latter depend on the accelerator you want
to use. Check :ref:`modules/model:Model` to see which ones are needed.


**Arguments:**

*--Optional--*


- **outfile_path** *(str)*:

    Name of fullresponse file.


- **creator** *(str)*:

    Create either with madx or analytically from twiss file.

    choices: ``('madx', 'twiss')``

    default: ``madx``


- **debug**:

    Print debug information.

    action: ``store_true``


- **delta_k** *(float)*:

    Delta K1 to be applied to quads for sensitivity matrix (madx-only).

    default: ``2e-05``


- **optics_params** *(str)*:

    List of parameters to correct upon (e.g. BBX BBY; twiss-only).


- **variable_categories**:

    List of the variables classes to use.

    default: ``['MQM', 'MQT', 'MQTL', 'MQY']``


"""
from typing import Dict

import pandas as pd
from generic_parser import DotDict
from generic_parser.entrypoint_parser import EntryPointParameters, entrypoint

from omc3.correction import response_madx, response_twiss
from omc3.correction.response_io import write_fullresponse
from omc3.global_correction import CORRECTION_DEFAULTS, OPTICS_PARAMS_CHOICES
from omc3.model import manager
from omc3.utils import logging_tools
from omc3.utils.iotools import PathOrStr

LOG = logging_tools.get_logger(__name__)


def response_params():
    params = EntryPointParameters()
    params.add_parameter(name="creator",
                         type=str,
                         choices=("madx", "twiss"),
                         default="madx",
                         help="Create either with madx or analytically from twiss file."
                         )
    params.add_parameter(name="variable_categories",
                         nargs="+",
                         default=CORRECTION_DEFAULTS["variable_categories"],
                         help="List of the variables classes to use."
                         )
    params.add_parameter(name="outfile_path",
                         type=PathOrStr,
                         help="Name of fullresponse file."
                         )
    params.add_parameter(name="delta_k",
                         type=float,
                         default=0.00002,
                         help="Delta K1 to be applied to quads for sensitivity matrix (madx-only).")
    params.add_parameter(name="optics_params",
                         type=str,
                         nargs="+",
                         choices=OPTICS_PARAMS_CHOICES,
                         help="List of parameters to correct upon (e.g. BBX BBY; twiss-only).",
                         )
    params.add_parameter(help="Print debug information.",
                         name="debug",
                         action="store_true",
                         )
    return params


@entrypoint(response_params())
def create_response_entrypoint(opt: DotDict, other_opt) -> Dict[str, pd.DataFrame]:
    """Entry point for creating pandas-based response matrices.

    The response matrices can be either created by response_madx or TwissResponse.
    """
    LOG.info("Creating response.")
    accel_inst = manager.get_accelerator(other_opt)

    if opt.creator == "madx":
        fullresponse = response_madx.create_fullresponse(
            accel_inst, opt.variable_categories, delta_k=opt.delta_k
        )

    elif opt.creator == "twiss":
        fullresponse = response_twiss.create_response(
            accel_inst, opt.variable_categories, opt.optics_params
        )

    if opt.outfile_path is not None:
        write_fullresponse(opt.outfile_path, fullresponse)
    return fullresponse


# Script Mode ------------------------------------------------------------------


if __name__ == "__main__":
    create_response_entrypoint()



"""
Interaction Point
-----------------

This module contains IP properties calculations related functionality of ``optics_measurements``.
It provides functions to compute beta* from phase.
"""
from os.path import join

import numpy as np
import pandas as pd
import tfs

from omc3.definitions.constants import PI2
from omc3.optics_measurements.constants import EXT, IP_NAME, S
from omc3.utils import logging_tools

LOGGER = logging_tools.get_logger(__name__)
COLUMNS = ("IP", "BETASTAR", "ERRBETASTAR", "PHASEADV", "ERRPHASEADV", "PHASEDVMDL", "LSTAR")


def betastar_from_phase(meas_input, phase_d):
    """
    Writes the **getIP** files with the betastar computed using phase advance.

    Arguments:
        meas_input: `Measurement_input` object.
        phase_d: Output of calculation.

    Returns:
        A nested `dict` with the same structure as the ``phase_d`` `dict`.
    """
    accel = meas_input.accelerator
    model = accel.model
    try:
        ips = list(accel.get_ips())
    except AttributeError:
        LOGGER.debug(f"Accelerator {type(accel).__name__} has not get_ips method.")
        return None
    rows = []
    for ip_name, bpml, bpmr in ips:
        try:
            phaseadv, ephaseadv, mdlphadv = _get_meas_phase(bpml, bpmr, phase_d)
        except KeyError:
            LOGGER.debug(f"{bpml} {bpmr} not in phases dataframe.")
            continue  # Measurement on one of the BPMs not present.
        lstar = _get_lstar(bpml, bpmr, model)
        betastar, ebestar = phase_to_betastar(lstar, PI2 * phaseadv, PI2 * ephaseadv)
        rows.append([ip_name, betastar, ebestar, phaseadv, ephaseadv, mdlphadv, lstar])
    return pd.DataFrame(columns=COLUMNS, data=rows)


def write(ips_d, headers, output_dir, plane):
    if ips_d is not None:
        tfs.write(join(output_dir, f"{IP_NAME}{plane.lower()}{EXT}"), ips_d, headers_dict=headers,)


def phase_to_betastar(lstar, phase, errphase):
    """
    Return the betastar and its error given the phase advance across the IP.

    This function computes the betastar using the phase advance between the BPMs around the IP
    and their distance (lstar). The phase and error in the phase must be given in radians.

    Args:
        lstar: The distance between the BPMs and the IR.
        phase: The phase advance between the BPMs at each side of the IP. Must be given in radians.
        errphase: The error in phase. Must be given in radians.
    """
    return (_phase_to_betastar_value(lstar, phase),
            _phase_to_betastar_error(lstar, phase, errphase))


def _phase_to_betastar_value(l, ph):
    tph = np.tan(ph)
    return (l * (1 - np.sqrt(tph ** 2 + 1))) / tph


def _phase_to_betastar_error(l, ph, eph):
    return abs((eph * l * (abs(np.cos(ph)) - 1)) / (np.sin(ph) ** 2))


def _get_meas_phase(bpml, bpmr, phases_df):
    return (phases_df["MEAS"].loc[bpml, bpmr],
            phases_df["ERRMEAS"].loc[bpml, bpmr],
            phases_df["MODEL"].loc[bpml, bpmr])


def _get_lstar(bpml, bpmr, model):
    return abs(model.loc[bpml, S] - model.loc[bpmr, S]) / 2.



"""
Stats
-----

Helper module providing statistical methods to compute various weighted averages along specified
axis and their errors as well as unbiased error estimator of infinite normal distribution from
finite-sized sample.

- TODO: use weighted average and its error in circular calculations
- TODO: write tests
- TODO: LOGGER or Raising error and warnings?
- TODO: if zeros or nans occur in errors, fallback to uniform weights only in affected cases
"""
import numpy as np
from scipy.special import erf
from scipy.stats import t
from omc3.definitions.constants import PI2, PI2I

CONFIDENCE_LEVEL = (1 + erf(1 / np.sqrt(2))) / 2


def circular_mean(data, period=PI2, errors=None, axis=None):
    """
    Computes weighted circular average along the specified axis.

    Parameters:
        data: array-like
            Contains the data to be averaged
        period: scalar, optional, default (2 * np.pi)
            Periodicity period of data
        errors: array-like, optional
            Contains errors associated with the values in data, it is used to calculated weights
        axis: int or tuple of ints, optional
            Axis or axes along which to average data

    Returns:
        Returns the weighted circular average along the specified axis.
    """

    phases = data * PI2I / period
    weights = weights_from_errors(errors, period=period)

    return np.angle(np.average(np.exp(phases), axis=axis, weights=weights)) * period / PI2


def circular_nanmean(data, period=PI2, errors=None, axis=None):
    """"Wrapper around circular_mean with added nan handling"""
    return circular_mean(data=np.ma.array(data, mask=np.isnan(data)),
                         period=period,
                         errors= None if errors is None else np.ma.array(errors, mask=np.isnan(data)),
                         axis=axis)


def circular_error(data, period=PI2, errors=None, axis=None, t_value_corr=True):
    """
    Computes error of weighted circular average along the specified axis.

    Parameters:
        data: array-like
            Contains the data to be averaged
        period: scalar, optional
            Periodicity period of data, default is (2 * np.pi)
        errors: array-like, optional
            Contains errors associated with the values in data, it is used to calculated weights
        axis: int or tuple of ints, optional
            Axis or axes along which to average data
        t_value_corr: bool, optional
            Species if the error is corrected for small sample size, default True

    Returns:
        Returns the error of weighted circular average along the specified axis.
        """
    phases = data * PI2I / period
    weights = weights_from_errors(errors, period=period)
    complex_phases = np.exp(phases)
    complex_average = np.average(complex_phases, axis=axis, weights=weights)

    (sample_variance, sum_of_weights) = np.average(
            np.square(np.abs(complex_phases - complex_average.reshape(_get_shape(
                    complex_phases.shape, axis)))), weights=weights, axis=axis, returned=True)
    if weights is not None:
        sample_variance = sample_variance + 1. / sum_of_weights
    error_of_complex_average = np.sqrt(sample_variance * unbias_variance(data, weights, axis=axis))
    phase_error = np.nan_to_num(error_of_complex_average / np.abs(complex_average))
    if t_value_corr:
        phase_error = phase_error * t_value_correction(effective_sample_size(data, weights, axis=axis))
    return np.where(phase_error > 0.25 * PI2, 0.3 * period, phase_error * period / PI2)


def circular_nanerror(data, period=PI2, errors=None, axis=None, t_value_corr=True):
    """"Wrapper around circular_error with added nan handling"""
    return circular_error(data=np.ma.array(data, mask=np.isnan(data)),
                          period=period,
                          errors=None if errors is None else np.ma.array(errors, mask=np.isnan(data)),
                          axis=axis,
                          t_value_corr=t_value_corr)


def weighted_mean(data, errors=None, axis=None):
    """
    Computes weighted average along the specified axis.

    Parameters:
        data: array-like
            Contains the data to be averaged
        errors: array-like, optional
            Contains errors associated with the values in data, it is used to calculated weights
        axis: int or tuple of ints, optional
            Axis or axes along which to average data

    Returns:
        Returns the weighted average along the specified axis.
    """
    weights = weights_from_errors(errors)
    return np.average(data, axis=axis, weights=weights)


def weighted_nanmean(data, errors=None, axis=None):
    """"Wrapper around weighted_mean with added nan handling"""
    return weighted_mean(data=np.ma.array(data, mask=np.isnan(data)),
                         errors=None if errors is None else np.ma.array(errors, mask=np.isnan(data)),
                         axis=axis)


def _get_shape(orig_shape, axis):
    new_shape = np.array(orig_shape)
    if axis is None:
        new_shape[:] = 1
    else:
        new_shape[np.array(axis)] = 1
    return tuple(new_shape)


def weighted_error(data, errors=None, axis=None, t_value_corr=True):
    """
    Computes error of weighted average along the specified axis.

    Parameters:
        data: array-like
            Contains the data to be averaged
        errors: array-like, optional
            Contains errors associated with the values in data, it is used to calculated weights
        axis: int or tuple of ints, optional
            Axis or axes along which to average data
        t_value_corr: bool, optional
            Species if the error is corrected for small sample size, default True

    Returns:
        Returns the error of weighted average along the specified axis.
    """
    weights = weights_from_errors(errors)
    weighted_average = np.average(data, axis=axis, weights=weights)
    (sample_variance, sum_of_weights) = np.ma.average(np.square(np.abs(data - weighted_average.reshape(
            _get_shape(data.shape, axis)))), weights=weights, axis=axis, returned=True)
    if weights is not None:
        sample_variance = sample_variance + 1 / sum_of_weights
    error = np.nan_to_num(np.sqrt(sample_variance * unbias_variance(data, weights, axis=axis)))
    if t_value_corr:
        error = error * t_value_correction(effective_sample_size(data, weights, axis=axis))
    return error


def weighted_rms(data, errors=None, axis=None):
    """
    Computes weighted root mean square along the specified axis.

    Parameters:
        data: array-like
            Contains the data to be averaged
        errors: array-like, optional
            Contains errors associated with the values in data, it is used to calculated weights
        axis: int or tuple of ints, optional
            Axis or axes along which to average data

    Returns:
        Returns weighted root mean square along the specified axis.
    """
    weights = weights_from_errors(errors)
    return np.sqrt(np.average(np.square(data), weights=weights, axis=axis))


def weighted_nanrms(data, errors=None, axis=None):
    """"Wrapper around weigthed_rms with added nan handling"""
    return weighted_rms(data=np.ma.array(data, mask=np.isnan(data)),
                        errors=None if errors is None else np.ma.array(errors, mask=np.isnan(data)),
                        axis=axis)


def weights_from_errors(errors, period=PI2):
    """
    Computes weights from measurement errors, weights are not output if errors contain zeros or nans

    Parameters:
        errors: array-like
            Contains errors which are used to calculated weights
        period: scalar, optional
            Periodicity period of data, default is (2 * np.pi)

    Returns:
        Returns the error of weighted circular average along the specified axis.
    """
    if errors is None:
        return None
    if np.any(np.isnan(errors)):
        # LOGGER.warning("Nans found, weights are not used.")
        return None
    if np.any(np.logical_not(errors)):
        # LOGGER.warning("Zeros found, weights are not used.")
        return None
    return 1 / np.square(errors * PI2 / period)


def effective_sample_size(data, weights, axis=None):
    """
    Computes effective sample size of weighted data along specifies axis,
    the minimum value returned is 2 to avoid non-reasonable error blow-up

    Parameters:
        data: array-like
        weights: array-like
            Contains weights associated with the values in data
        axis: int or tuple of ints, optional
            Axis or axes along which the effective sample size is computed

    Returns:
        Returns the error of weighted circular average along the specified axis.
    """
    if weights is None:
        sample_size = np.sum(np.ones(data.shape), axis=axis)
    else:
        sample_size = np.square(np.sum(weights, axis=axis)) / np.sum(np.square(weights), axis=axis)
    return np.where(sample_size > 2, sample_size, 2)


def unbias_variance(data, weights, axis=None):
    """
    Computes a correction factor to unbias variance of weighted average of data along specified axis

    Parameters:
        data: array-like
        weights: array-like
            Contains weights associated with the values in data
        axis: int or tuple of ints, optional
            Axis or axes along which the effective sample size is computed

    Returns:
        Returns the error of weighted circular average along the specified axis.
    """
    sample_size = effective_sample_size(data, weights, axis=axis)
    try:
        return sample_size / (sample_size - 1)
    except ZeroDivisionError or RuntimeWarning:
        return np.zeros(sample_size.shape)


def t_value_correction(sample_size):
    """
    Calculates the multiplicative correction factor to determine standard deviation of normally
    distributed quantity from standard deviation of its finite-sized sample
    the minimum allowed sample size is 2 to avoid non-reasonable error blow-up
    for smaller sample sizes 2 is used instead

    Args:
        sample_size: array-like

    Returns:
        multiplicative correction factor(s) of same shape as sample_size
            can contain nans
    """
    return t.ppf(CONFIDENCE_LEVEL, np.where(sample_size > 2, sample_size, 2) - 1)



"""
Constants
---------

Specific constants to be used in ``harpy``, to help with consistency.
"""
# Output Fileextensions --------------------------------------------------------

FILE_AMPS_EXT = ".amps{plane}"
FILE_FREQS_EXT = ".freqs{plane}"
FILE_LIN_EXT = ".lin{plane}"

# Column Names -----------------------------------------------------------------
# TODO use these everywhere (jdilly)

# Basic ---
COL_NAME = "NAME"

# Lin Files ---
COL_TUNE = "TUNE"
COL_AMP = "AMP"
COL_MU = "MU"

COL_NATTUNE = "NATTUNE"
COL_NATAMP = "NATAMP"
COL_NATMU = "NATMU"

COL_FREQ = "FREQ"
COL_PHASE = "PHASE"
COL_ERR = "ERR"



"""
IO Tools
--------

Helper functions for input/output issues.
"""
import json
import os
import shutil
from pathlib import Path

from generic_parser.entry_datatypes import get_instance_faker_meta
from generic_parser.entrypoint_parser import save_options_to_config
from pandas import DataFrame
from tfs import TfsDataFrame

from omc3.definitions import formats
from omc3.utils import logging_tools

LOG = logging_tools.get_logger(__name__)


def copy_content_of_dir(src_dir, dst_dir):
    """Copies all files and directories from ``src_dir`` to ``dst_dir``."""
    if not os.path.isdir(src_dir):
        return

    create_dirs(dst_dir)

    for item in os.listdir(src_dir):
        src_item = os.path.join(src_dir, item)
        dst_item = os.path.join(dst_dir, item)
        copy_item(src_item, dst_item)


def create_dirs(path_to_dir):
    """Creates all dirs to ``path_to_dir`` if not exists."""
    path_to_dir = Path(path_to_dir)
    if not path_to_dir.exists():
        path_to_dir.mkdir(parents=True)
        LOG.debug("Created directory structure: '{}'".format(path_to_dir))


def copy_item(src_item, dest):
    """
    Copies a file or a directory to ``dest``, which may be a directory.
    If ``src_item`` is a directory then all containing files and dirs will be copied into ``dest``.
    """
    try:
        if os.path.isfile(src_item):
            shutil.copy2(src_item, dest)
        elif os.path.isdir(src_item):
            copy_content_of_dir(src_item, dest)
        else:
            raise IOError
    except IOError:
        LOG.error("Could not copy item because of IOError. Item: '{}'".format(src_item))


class PathOrStr(metaclass=get_instance_faker_meta(Path, str)):
    """A class that behaves like a Path when possible, otherwise like a string."""
    def __new__(cls, value):
        if isinstance(value, str):
            value = value.strip("\'\"")  # behavior like dict-parser, IMPORTANT FOR EVERY STRING-FAKER
        return Path(value)


class PathOrStrOrDataFrame(metaclass=get_instance_faker_meta(TfsDataFrame, Path, str)):
    """A class that behaves like a Path when possible, otherwise like a string."""
    def __new__(cls, value):
        if isinstance(value, str):
            value = value.strip("\'\"")  # behavior like dict-parser, IMPORTANT FOR EVERY STRING-FAKER
        try:
            return Path(value)
        except TypeError:
            TfsDataFrame(value)


class UnionPathStr(metaclass=get_instance_faker_meta(Path, str)):
    """A class that can be used as Path and string parser input, but does not convert."""
    def __new__(cls, value):
        if isinstance(value, str):
            value = value.strip("\'\"")  # behavior like dict-parser, IMPORTANT FOR EVERY STRING-FAKER
        return value


class UnionPathStrInt(metaclass=get_instance_faker_meta(Path, str, int)):
    """A class that can be used as Path, string, int parser input, but does not convert.
    Very special and used e.g. in the BBQ Input."""
    def __new__(cls, value):
        if isinstance(value, str):
            value = value.strip("\'\"")  # behavior like dict-parser, IMPORTANT FOR EVERY STRING-FAKER
        return value


def convert_paths_in_dict_to_strings(dict_: dict) -> dict:
    """Converts all Paths in the dict to strings, including those in iterables."""
    dict_ = dict_.copy()
    for key, value in dict_.items():
        if isinstance(value, Path):
            dict_[key] = str(value)
        else:
            try:
                list_ = list(value)
            except TypeError:
                pass
            else:
                has_changed = False
                for idx, item in enumerate(list_):
                    if isinstance(item, Path):
                        list_[idx] = str(item)
                        has_changed = True
                if has_changed:
                    dict_[key] = list_
    return dict_


def remove_none_dict_entries(dict_: dict) -> dict:
    """
    Removes ``None`` entries from dict. This can be used as a workaround to
    https://github.com/pylhc/generic_parser/issues/26.
    """
    return {key: value for key, value in dict_.items() if value is not None}


def save_config(output_dir: Path, opt: dict, script: str):
    """
    Quick wrapper for ``save_options_to_config``.

    Args:
        output_dir (Path): Path to the output directory (does not need to exist).
        opt (dict): opt-structure to be saved.
        script (str): path/name of the invoking script (becomes name of the .ini) usually
            ``__file__``.
    """
    output_dir.mkdir(parents=True, exist_ok=True)
    opt = remove_none_dict_entries(opt)  # temporary fix (see docstring)
    opt = convert_paths_in_dict_to_strings(opt)
    save_options_to_config(output_dir / formats.get_config_filename(script),
                           dict(sorted(opt.items()))
                           )



"""
Constants
---------

Specific constants and helpers to be used in ``tune_analysis``, to help with consistency.
"""
from dataclasses import dataclass
from typing import Tuple, Sequence

import pandas as pd
from omc3.definitions.constants import PLANE_TO_NUM
from omc3.optics_measurements.constants import ACTION, ERR, EXT, KICK_NAME, NAT_TUNE, RES, TIME

# Global 'Parameters' for easy editing #########################################

ODR_PREF: str = "ODR_"
MOVING_AV: str = "MAV"
TOTAL: str = "TOT"
CORRECTED: str = "CORR"
COEFFICIENT: str = "COEFF{order:d}"
BBQ: str = "BBQ"

INPUT_KICK = "kick"
INPUT_PREVIOUS = "previous"


def get_timber_bbq_key(plane: str, beam: int) -> str:
    """ Key to extract bbq from timber. """
    # return f"lhc.bofsu:eigen_freq_{PLANE_TO_NUM[plane]:d}_b{beam:d}".upper()  # pre run 3
    # return f"BFC.LHC:TuneFBAcq:TUNEB{beam:d}{PLANE_TO_HV[plane]}"  # contains less data
    return f"LHC.BQBBQ.CONTINUOUS_HS.B{beam:d}:EIGEN_FREQ_{PLANE_TO_NUM[plane]:d}"


def get_kick_out_name() -> str:
    return f"{KICK_NAME}ampdet_xy{EXT}"


def get_bbq_out_name() -> str:
    return f"bbq_ampdet.tfs"


@dataclass
class AmpDetData:
    tune_plane: str
    action_plane: str
    action: pd.Series
    action_err: pd.Series
    tune: pd.Series
    tune_err: pd.Series


@dataclass
class FakeOdrOutput:
    beta: Sequence[float]
    sd_beta: Sequence[float]


# Kick File Headers ###########################################################


def get_tstart_head() -> str:
    """ Label for fill start time from header. """
    return "START_TIME"


def get_tend_head() -> str:
    """ Label for fill end time from header. """
    return "END_TIME"


def get_odr_header_default(q_plane: str, j_plane: str) -> str:
    return f"{ODR_PREF}dQ{q_plane.upper():s}d2J{j_plane.upper():s}"


def get_odr_header_coeff(q_plane: str, j_plane: str, order: int) -> str:
    """ Header key for odr coefficient for term of given order (i.e. beta[order]) """
    return f"{get_odr_header_default(q_plane, j_plane) :s}_{COEFFICIENT.format(order=order)}"


def get_odr_header_err_coeff(q_plane: str, j_plane: str, order: int) -> str:
    """ Header key for odr coefficient standard deviation for term of given order (i.e. sd_beta[order]) """
    return f"{get_odr_header_default(q_plane, j_plane) :s}_{ERR}{COEFFICIENT.format(order=order)}"


def get_odr_header_coeff_corrected(q_plane: str, j_plane: str, order: int) -> str:
    """ Header key for corrected odr coefficient for term of given order (i.e. beta[order]) """
    return f"{get_odr_header_default(q_plane, j_plane)}_{CORRECTED}{COEFFICIENT.format(order=order)}"


def get_odr_header_err_coeff_corrected(q_plane: str, j_plane: str, order: int) -> str:
    """ Header key for corrected odr coefficient standard deviation for term of given order (i.e. sd_beta[order]) """
    return f"{get_odr_header_default(q_plane, j_plane)}_{ERR}{CORRECTED}{COEFFICIENT.format(order=order)}"


def get_mav_window_header() -> str:
    """ Header to store the moving average window length """
    return "MOVINGAV_WINDOW"


def get_fine_window_header() -> str:
    """ Header to store the moving average fine window length """
    return "FINE_WINDOW"


def get_fine_cut_header() -> str:
    """ Header to store the fine cut. """
    return "FINE_CUT"


def get_outlier_limit_header() -> str:
    """ Header to store the outlier limit."""
    return "OUTLIER_LIMIT"


def get_min_tune_header(plane: str) -> str:
    """ Header to store the minimum tune cut. """
    return f"MIN_Q{plane.upper()}"


def get_max_tune_header(plane: str) -> str:
    """ Header to store the maximum tune cut. """
    return f"MAX_Q{plane.upper()}"


# Kick File Columns ###########################################################


def get_time_col() -> str:
    """ Label for the TIME column."""
    return TIME


def get_bbq_col(plane: str) -> str:
    """ Label for the BBQ column """
    return f"{BBQ}{plane.upper():s}"


def get_mav_col(plane: str) -> str:
    """ Label for the moving average BBQ column. """
    return f"{get_bbq_col(plane):s}{MOVING_AV}"


def get_used_in_mav_col(plane: str) -> str:
    """ Label for the column showing if BBQ value was used in moving average. """
    return f"{get_bbq_col(plane):s}IN{MOVING_AV}"


def get_mav_err_col(plane: str) -> str:
    """ Label for the standard deviation of the moving average data. """
    return f"{ERR}{get_bbq_col(plane):s}{MOVING_AV}"


def get_corr_natq_err_col(plane: str) -> str:
    """ Return the standard deviation for the corrected natural tune. """
    return f"{ERR}{get_natq_corr_col(plane):s}"


def get_natq_col(plane: str) -> str:
    """ Label for the natural tune column. """
    return f"{NAT_TUNE}{plane.upper():s}"


def get_natq_corr_col(plane: str) -> str:
    """ Label for the corrected natural tune column. """
    return f"{get_natq_col(plane):s}{CORRECTED}"


def get_natq_err_col(plane: str) -> str:
    """ Label for the natural tune error column. """
    return f"{ERR}{get_natq_col(plane):s}"


def get_action_col(plane: str) -> str:
    """ Label for the action column. """
    return f"{ACTION}{plane.upper():s}{RES}"


def get_action_err_col(plane: str) -> str:
    """ Label for the action error column. """
    return f"{ERR}{get_action_col(plane):s}"


# Plotting #####################################################################


def get_tune_label(plane: str, scale: int = None) -> str:
    """ Tune label for the action/tune plots. """
    unit = ""
    if scale:
        unit = f" \quad [10^{{{scale:d}}}]"
    return fr"$\Delta Q_{plane.lower():s}{unit}$"


def get_action_label(plane: str, unit: str) -> str:
    """ Action label for the action/tune plots. """
    if unit == "um":
        unit = r"\mu m"
    return fr"$2J_{plane.lower():s} \quad [{unit:s}]$"


def get_detuning_exponent_for_order(order: int) -> int:
    """ Returns the default exponent for detuning orders. Highly Empirical. """
    return {1: 3, 2: 9, 3: 15}[order]



"""
Kick File Modifiers
-------------------

Functions to add data to or extract data from **kick_ac** files.
"""
import os
from pathlib import Path
from typing import Tuple, Union, Sequence

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray
from scipy import odr
import tfs
from tfs import TfsDataFrame
from generic_parser.dict_parser import DotDict

from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import KICK_NAME, TIME, ERR, NAT_TUNE, ACTION
from omc3.tune_analysis import bbq_tools
from omc3.tune_analysis.bbq_tools import OutlierFilterOpt, FilterOpts, MinMaxFilterOpt
from omc3.tune_analysis.constants import (get_odr_header_coeff_corrected,
                                          get_odr_header_err_coeff_corrected,
                                          get_odr_header_err_coeff, get_odr_header_coeff,
                                          get_natq_corr_col, get_corr_natq_err_col,
                                          get_natq_err_col, get_natq_col, get_bbq_col,
                                          get_mav_col, get_mav_err_col, get_used_in_mav_col,
                                          get_time_col, get_action_col, get_action_err_col,
                                          get_mav_window_header, get_outlier_limit_header,
                                          get_fine_window_header, get_fine_cut_header,
                                          get_min_tune_header, get_max_tune_header,
                                          AmpDetData, FakeOdrOutput, get_kick_out_name, COEFFICIENT, CORRECTED,
                                          get_odr_header_default,
                                          )
from omc3.utils import logging_tools
from omc3.utils.time_tools import CERNDatetime, get_cern_time_format

LOG = logging_tools.get_logger(__name__)


def _get_odr_headers(corrected):
    """Return Headers needed for ODR."""
    if corrected:
        return get_odr_header_coeff_corrected, get_odr_header_err_coeff_corrected
    return get_odr_header_coeff, get_odr_header_err_coeff


def _get_ampdet_columns(corrected):
    """Get columns needed for amplitude detuning."""
    if corrected:
        return get_natq_corr_col, get_corr_natq_err_col
    return get_natq_col, get_natq_err_col


# Data Addition ################################################################


def add_bbq_data(kick_df: pd.DataFrame, bbq_df: pd.DataFrame, column: str, bbq_column: str = None):
    """
    Add BBQ values from column to kickac dataframe into same column.

    Args:
        kick_df: kick `DataFrame`, which needs to have time as index, best load it with
            ``read_timed_dataframe()``.
        bbq_df: `DataFrame` of bbq data with time as index.
        column: column name to add the data into.
        bbq_column: column name to get the data from (same as `column` if not given).

    Returns:
        Modified kick `Dataframe`.
    """
    if bbq_column is None:
        bbq_column = column
    kick_indx = get_timestamp_index(kick_df.index)
    bbq_indx = get_timestamp_index(bbq_df.index)

    values = []
    for time in kick_indx:
        nearest_bbq_index = bbq_indx.get_indexer([time], method="nearest")[0]
        values.append(bbq_df[bbq_column].iloc[nearest_bbq_index])
    kick_df[column] = values
    return kick_df


def add_moving_average(kickac_df: TfsDataFrame, bbq_df: TfsDataFrame, filter_opt: FilterOpts,
                       ) -> Tuple[TfsDataFrame, TfsDataFrame]:
    """Adds the moving average of the bbq data to kickac_df and bbq_df."""
    LOG.debug("Calculating moving average.")

    if isinstance(filter_opt, OutlierFilterOpt):
        bbq_df = _filter_bbq_outliers(bbq_df, filter_opt)
    else:
        bbq_df = _filter_bbq_cut(bbq_df, filter_opt)

    for plane in PLANES:
        kickac_df = add_bbq_data(kickac_df, bbq_df, get_mav_col(plane))
        kickac_df = add_bbq_data(kickac_df, bbq_df, get_mav_err_col(plane))
    return kickac_df, bbq_df


def _filter_bbq_outliers(bbq_df: tfs.TfsDataFrame, filter_opt: OutlierFilterOpt) -> tfs.TfsDataFrame:
    header_limit = get_outlier_limit_header()
    header_window = get_mav_window_header()

    # check if we need to recompute (as this might take quite some time)
    if (
            (header_limit in bbq_df.headers)
            and (bbq_df.headers[header_limit] == filter_opt.limit)
            and (bbq_df.headers[header_window] == filter_opt.window)
            and all(col in bbq_df.columns for p in PLANES for col in (
                    get_mav_col(p), get_mav_err_col(p), get_used_in_mav_col(p)))
    ):
        LOG.info("BBQ data has already been filtered with the same parameters. Using data from file.")
        return bbq_df

    bbq_df.headers[header_window] = filter_opt.window
    bbq_df.headers[header_limit] = filter_opt.limit
    for plane in PLANES:
        bbq_mav, bbq_std, mask = bbq_tools.clean_outliers_moving_average(bbq_df[get_bbq_col(plane)],
                                                                         filter_opt=filter_opt)
        bbq_df[get_mav_col(plane)] = bbq_mav
        # bbq_df[get_mav_err_col(plane)] = bbq_std  # this is too large
        bbq_df[get_mav_err_col(plane)] = 0.  # TODO to be discussed with Ewen and Tobias (jdilly, 2022-05-23)
        bbq_df[get_used_in_mav_col(plane)] = mask
    return bbq_df


def _filter_bbq_cut(bbq_df: tfs.TfsDataFrame, filter_opts: Sequence[MinMaxFilterOpt]) -> tfs.TfsDataFrame:
    bbq_df.headers[get_mav_window_header()] = filter_opts[0].window
    bbq_df.headers[get_fine_window_header()] = filter_opts[0].fine_window
    bbq_df.headers[get_fine_cut_header()] = filter_opts[0].fine_cut

    for idx, plane in enumerate(PLANES):
        bbq_mav, bbq_std, mask = bbq_tools.get_moving_average(bbq_df[get_bbq_col(plane)], filter_opts[idx])

        bbq_df.headers[get_min_tune_header(plane)] = filter_opts[idx].min
        bbq_df.headers[get_max_tune_header(plane)] = filter_opts[idx].max

        bbq_df[get_mav_col(plane)] = bbq_mav
        # bbq_df[get_mav_err_col(plane)] = bbq_std  # this is too large
        bbq_df[get_mav_err_col(plane)] = 0.  # TODO to be discussed with Ewen and Tobias (jdilly, 2022-05-23)
        bbq_df[get_used_in_mav_col(plane)] = mask
    return bbq_df


def add_corrected_natural_tunes(kickac_df: pd.DataFrame) -> pd.DataFrame:
    """
    Adds the corrected natural tunes to ``kickac_df``.
    Add also the total standard deviation of the natural tune to the kickac dataframe.
    The total standard deviation is here defined as the standard deviation of
    the measurement plus the standard deviation of the moving average.

    Args:
        kickac_df: `Dataframe` containing the data.

    Returns:
        Modified kick_ac.
    """
    for plane in PLANES:
        # NATQ
        kickac_df[get_natq_corr_col(plane)] = (
            kickac_df[get_natq_col(plane)] - kickac_df[get_mav_col(plane)]
        )

        # ERROR
        kickac_df[get_corr_natq_err_col(plane)] = np.sqrt(
            np.power(kickac_df[get_natq_err_col(plane)], 2) +
            np.power(kickac_df[get_mav_err_col(plane)], 2)
        )
    return kickac_df


def add_odr(kickac_df: pd.DataFrame, odr_fit: odr.Output, 
            action_plane: str, tune_plane: str, corrected: bool = False):
    """
    Adds the odr fit of the (un)corrected data to the header of the ``kickac_df``.

    Args:
        kickac_df: `Dataframe` containing the data.
        odr_fit: odr-fit data (definitions see ``detuning_tools.py``).
        action_plane: Plane of the action.
        tune_plane: Plane of the tune.
        corrected: (BBQ) corrected data or uncorrected fit?.

    Returns:
        Modified kick_ac.
    """
    header_val, header_err = _get_odr_headers(corrected)
    for idx in range(len(odr_fit.beta)):
        kickac_df.headers[header_val(j_plane=action_plane, q_plane=tune_plane, order=idx)] = odr_fit.beta[idx]
        kickac_df.headers[header_err(j_plane=action_plane, q_plane=tune_plane, order=idx)] = odr_fit.sd_beta[idx]
    return kickac_df


# Data Extraction --------------------------------------------------------------

def get_odr_data(kickac_df: pd.DataFrame, action_plane: str, tune_plane: str,
                 order: int, corrected: bool = False) -> FakeOdrOutput:
    """
    Extract the data from kickac.

    Args:
        kickac_df: `Dataframe` containing the data.
        action_plane: Plane of the action.
        tune_plane: Plane of the tune.
        order: Order of the odr fit.
        corrected: (BBQ) corrected data or uncorrected fit?

    Returns:
        `Dictionary` containing fit data from odr.
    """

    header_val, header_err = _get_odr_headers(corrected)
    odr_data = FakeOdrOutput(beta=[0] * (order+1), sd_beta=[0] * (order+1))
    for idx in range(order+1):
        odr_data.beta[idx] = kickac_df.headers[header_val(q_plane=tune_plane, j_plane=action_plane, order=idx)]
        odr_data.sd_beta[idx] = kickac_df.headers[header_err(q_plane=tune_plane, j_plane=action_plane, order=idx)]
    return odr_data


def get_ampdet_data(kickac_df: pd.DataFrame, action_plane: str, tune_plane: str,
                    corrected: bool = False, dropna: bool = True) -> AmpDetData:
    """
    Extract the data needed for the (un)corrected amplitude detuning from ``kickac_df``.

    Args:
        kickac_df (DataFrame): Containing the data
                              (either `kick_x`, `kick_y` from optics
                              or `kick_ampdet_xy` from ampdet analysis)..
        action_plane (str): Plane of the action.
        tune_plane (str): Plane of the tune.
        corrected (bool): if the BBQ-corrected columns should be taken
        dropna (bool): drop columns containing NaNs

    Returns:
        `Dataframe` containing `action`, `tune`, `action_err` and `tune_err` columns.
    """
    col_natq, col_natq_std = _get_ampdet_columns(corrected)

    columns = {"action": get_action_col(action_plane),
               "action_err": get_action_err_col(action_plane),
               "tune": col_natq(tune_plane),
               "tune_err": col_natq_std(tune_plane),
               }

    not_found = [cv for cv in columns.values() if cv not in kickac_df.columns]
    if any(not_found):
        raise KeyError(f"The following columns were not found in kick-file: '{str(not_found)}'")

    data = kickac_df.loc[:, list(columns.values())]
    data.columns = columns.keys()

    if data.isna().any().any():
        # could be on purpose,
        # for manually filtering kick points in one plane, but not in the other
        LOG.warning(
            f"Amplitude Detuning data for Q{tune_plane} and J{action_plane} contains NaNs"
        )
        if dropna:
            LOG.debug("NaN's are being dropped.")
            data = data.dropna(axis=0)

    return AmpDetData(
        action_plane=action_plane,
        tune_plane=tune_plane,
        **{column: data[column] for column in data.columns}
    )


# Timed DataFrames -------------------------------------------------------------


def get_timestamp_index(index: pd.Index) -> pd.Index:
    return pd.Index([i.timestamp() for i in index])


def read_timed_dataframe(path: Union[str, Path]) -> tfs.TfsDataFrame:
    df = tfs.read(path, index=get_time_col())
    df.index = pd.Index([CERNDatetime.from_cern_utc_string(i) for i in df.index], dtype=object)
    return df


def write_timed_dataframe(path: Union[str, Path], df: tfs.TfsDataFrame):
    df = df.copy()
    df.index = pd.Index([i.strftime(get_cern_time_format()) for i in df.index], dtype=str)
    tfs.write(path, df, save_index=get_time_col(), headerswidth=max(len(k) for k in df.headers.keys()))


def read_two_kick_files_from_folder(folder: Path) -> tfs.TfsDataFrame:
    return merge_two_plane_kick_dfs(
        *[read_timed_dataframe(folder / f'{KICK_NAME}{p.lower()}.tfs') for p in PLANES]
    )


def merge_two_plane_kick_dfs(df_x: tfs.TfsDataFrame, df_y: tfs.TfsDataFrame) -> tfs.TfsDataFrame:
    df_xy = TfsDataFrame(pd.merge(df_x, df_y, how='outer', left_index=True, right_index=True, suffixes=PLANES))
    if len(df_xy.index) != len(df_x.index) or len(df_xy.index) != len(df_y.index):
        LOG.warning("The kick-files for each plane seem to have different indices.")
    df_xy.headers = df_x.headers
    df_xy.headers.update(df_y.headers)
    return df_xy


# BBS Converter ----------------------------------------------------------------

def convert_bbs_kickfile(file: Path):
    """ Wrapper for :meth:`omc3.tune_analysis.kick_file_modifiers.convert_old_kickdataframe`
    with read/write capabilities."""
    df = tfs.read(file)
    df = convert_bbs_kickdataframe(df)
    tfs.write(file.with_name(get_kick_out_name()), df)


def convert_bbs_kickdataframe(df: tfs.TfsDataFrame) -> tfs.TfsDataFrame:
    """Converts a kick-dataframe created with the python2 amplitude detuning
    analysis into a dataframe compatible with omc3's detuning analysis.
    Beware that the old files may not contain the error coefficients for
    the corrected data in the header (the uncorrected errors are used here),
    as well as no corrected natural tune error columns.
    And that the offset (coeffient 0) has no error, which is here assumed
    to be zero (it is not used anyway, but required when reading the data).
    The action unit is converted from um to m.
    !!! Be also aware, that due to a bug the Action and Tune planes might be switched !!!
    This is not taken into account here.
    """
    # Header ---
    df.headers = {_rename_old_header(key): value for key, value in df.headers.items()}
    for key in list(df.headers.keys()):
        # convert coeffients (only first order was implemented in python2)
        if COEFFICIENT.format(order=1) in key:
            df.headers[key] = df.headers[key] * 1e6  # inverse um to inverse m

        # use uncorrected error columns for corrected data
        if ERR in key:
            new_key = key.replace(ERR, f"{ERR}{CORRECTED}")
            df.headers[new_key] = df.headers[key]

    # add err headers for coefficient 0 (offset)
    for tune in PLANES:
        for action in PLANES:
            name = get_odr_header_err_coeff(q_plane=tune, j_plane=action, order=0)
            df.headers[name] = df.headers.get(name, 0)

            name = get_odr_header_err_coeff_corrected(q_plane=tune, j_plane=action, order=0)
            df.headers[name] = df.headers.get(name, 0)

    # Columns ---
    df[TIME] = [CERNDatetime.from_timestamp(ts).cern_utc_string() for ts in df[TIME]]
    df.columns = [_rename_old_column(c) for c in df.columns]
    for column in df.columns:
        if ACTION in column:
            df[column] = df[column] * 1e-6  # um to m

    for plane in PLANES:
        col = f"{ERR}{NAT_TUNE}{plane}{CORRECTED}"
        if col not in df.columns:
            df[col] = df[f"{ERR}{NAT_TUNE}{plane}"]
    return df


def _rename_old_header(key: str):
    for tune in PLANES:
        for action in PLANES:
            old = f"ODR_J{action}Q{tune}"
            new = get_odr_header_default(q_plane=tune, j_plane=action)
            key = key.replace(old, new)
    key = key.replace("OFFSET", COEFFICIENT.format(order=0)).replace("SLOPE", COEFFICIENT.format(order=1))

    # this was the last prefix in the old file, so replace this first
    if "_CORR" in key:
        parts = key.split("_")
        key = "_".join(parts[:-2] + [f"{CORRECTED}{parts[-2]}"])

    # and then remove also this and add ERR
    if "_STD" in key:
        parts = key.split("_")
        key = "_".join(parts[:-2] + [f"{ERR}{parts[-2]}"])
    return key


def _rename_old_column(column: str):
    if "STD" in column or "RMS" in column:
        return f"{ERR}{column.replace('STD', '').replace('RMS', '')}"
    return column



"""
LHC Model Creator
-----------------

This module provides convenience functions for model creation of the ``LHC``.
"""
import logging
import shutil
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import tfs
from omc3.model.accelerators.accelerator import AcceleratorDefinitionError, AccExcitationMode
from omc3.model.accelerators.lhc import Lhc
from omc3.model.constants import (
    B2_ERRORS_TFS,
    B2_SETTINGS_MADX,
    ERROR_DEFFS_TXT,
    GENERAL_MACROS,
    LHC_MACROS,
    LHC_MACROS_RUN3,
    MACROS_DIR,
    TWISS_AC_DAT,
    TWISS_ADT_DAT,
    TWISS_BEST_KNOWLEDGE_DAT,
    TWISS_DAT,
    TWISS_ELEMENTS_BEST_KNOWLEDGE_DAT,
    TWISS_ELEMENTS_DAT,
    ACCELERATOR_MODEL_REPOSITORY
)
from omc3.model.model_creators.abstract_model_creator import ModelCreator
from omc3.utils import iotools

LHC_REPOSITORY_NAME = "acc-models-lhc"

LOGGER = logging.getLogger(__name__)


def _b2_columns() -> List[str]:
    cols_outer = [f"{KP}{num}{S}L" for KP in ("K", "P") for num in range(21) for S in ("", "S")]
    cols_middle = ["DX", "DY", "DS", "DPHI", "DTHETA", "DPSI", "MREX", "MREY", "MREDX", "MREDY",
                   "AREX", "AREY", "MSCALX", "MSCALY", "RFM_FREQ", "RFM_HARMON", "RFM_LAG"]
    return cols_outer[:42] + cols_middle + cols_outer[42:]


class LhcModelCreator(ModelCreator):
    @classmethod
    def get_madx_script(cls, accel: Lhc) -> str:  # nominal
        use_acd = "1" if (accel.excitation == AccExcitationMode.ACD) else "0"
        use_adt = "1" if (accel.excitation == AccExcitationMode.ADT) else "0"
        madx_script = accel.get_base_madx_script()
        madx_script += (
            f"exec, do_twiss_monitors(LHCB{accel.beam}, '{accel.model_dir / TWISS_DAT}', {accel.dpp});\n"
            f"exec, do_twiss_elements(LHCB{accel.beam}, '{accel.model_dir / TWISS_ELEMENTS_DAT}', {accel.dpp});\n"
        )
        if accel.excitation != AccExcitationMode.FREE or accel.drv_tunes is not None:
            # allow user to modify script and enable excitation, if driven tunes are given
            madx_script += (
                f"use_acd={use_acd};\nuse_adt={use_adt};\n"
                f"if(use_acd == 1){{\n"
                f"exec, twiss_ac_dipole({accel.nat_tunes[0]}, {accel.nat_tunes[1]}, {accel.drv_tunes[0]}, {accel.drv_tunes[1]}, {accel.beam}, '{accel.model_dir / TWISS_AC_DAT}', {accel.dpp});\n"
                f"}}else if(use_adt == 1){{\n"
                f"exec, twiss_adt({accel.nat_tunes[0]}, {accel.nat_tunes[1]}, {accel.drv_tunes[0]}, {accel.drv_tunes[1]}, {accel.beam}, '{accel.model_dir / TWISS_ADT_DAT}', {accel.dpp});\n"
                f"}}\n"
            )
        return madx_script

    @classmethod
    def get_correction_check_script(
        cls, accel: Lhc, corr_file: str = "changeparameters_couple.madx", chrom: bool = False
    ) -> str:
        madx_script = accel.get_base_madx_script()
        madx_script += (
            f"exec, do_twiss_monitors_and_ips(LHCB{accel.beam}, '{accel.model_dir / 'twiss_no.dat'!s}', 0.0);\n"
            f"call, file = '{corr_file}';\n"
            f"exec, do_twiss_monitors_and_ips(LHCB{accel.beam}, '{accel.model_dir / 'twiss_cor.dat'!s}', 0.0);\n"
        )
        if chrom:
            madx_script += (
                f"exec, do_twiss_monitors_and_ips(LHCB{accel.beam}, '{accel.model_dir / 'twiss_cor_dpm.dat'}', %DELTAPM);\n"
                f"exec, do_twiss_monitors_and_ips(LHCB{accel.beam}, '{accel.model_dir / 'twiss_cor_dpp.dat'}', %DELTAPP);\n"
            )
        return madx_script

    @classmethod
    def prepare_run(cls, accel: Lhc) -> None:
        if accel.year in ["2018", "2022"]:  # these years should be handled by the fetcher
            symlink_dst = Path(accel.model_dir)/LHC_REPOSITORY_NAME
            if not symlink_dst.exists():
                LOGGER.debug(f"Symlink destination: {symlink_dst}")
                symlink_dst.absolute().symlink_to((ACCELERATOR_MODEL_REPOSITORY/f"{accel.year}"))

        cls.check_accelerator_instance(accel)
        LOGGER.debug("Preparing model creation structure")
        macros_path = accel.model_dir / MACROS_DIR
        iotools.create_dirs(macros_path)

        LOGGER.debug("Copying macros to model directory")
        lib_path = Path(__file__).parent.parent / "madx_macros"
        shutil.copy(lib_path / GENERAL_MACROS, macros_path / GENERAL_MACROS)
        shutil.copy(lib_path / LHC_MACROS, macros_path / LHC_MACROS)
        shutil.copy(lib_path / LHC_MACROS_RUN3, macros_path / LHC_MACROS_RUN3)

        if accel.energy is not None:
            LOGGER.debug("Copying B2 error files for given energy in model directory")
            core = f"{int(accel.energy * 1000):04d}"
            error_dir_path = accel.get_lhc_error_dir()
            shutil.copy(error_dir_path / f"{core}GeV.tfs", accel.model_dir / ERROR_DEFFS_TXT)
            shutil.copy(
                error_dir_path / "b2_errors_settings" / f"beam{accel.beam}_{core}GeV.madx",
                accel.model_dir / B2_SETTINGS_MADX,
            )
            b2_table = tfs.read(error_dir_path / f"b2_errors_beam{accel.beam}.tfs", index="NAME")
            gen_df = pd.DataFrame(
                data=np.zeros((b2_table.index.size, len(_b2_columns()))),
                index=b2_table.index,
                columns=_b2_columns(),
            )
            gen_df["K1L"] = b2_table.loc[:, f"K1L_{core}"].to_numpy()
            tfs.write(
                accel.model_dir / B2_ERRORS_TFS,
                gen_df,
                headers_dict={"NAME": "EFIELD", "TYPE": "EFIELD"},
                save_index="NAME",
            )

    @staticmethod
    def check_accelerator_instance(accel: Lhc) -> None:
        accel.verify_object()  # should have been done anyway, but cannot hurt (jdilly)

        # Creator specific checks
        if accel.model_dir is None:
            raise AcceleratorDefinitionError(
                "The accelerator definition is incomplete: model directory (outputdir option) was not given."
            )

        if accel.modifiers is None or not len(accel.modifiers):
            raise AcceleratorDefinitionError(
                "The accelerator definition is incomplete: no modifiers could be found."
            )

        # hint: if modifiers are given as absolute paths: `path / abs_path` returns `abs_path`  (jdilly)
        inexistent_modifiers = [m for m in accel.modifiers if not (accel.model_dir / m).exists()]
        if len(inexistent_modifiers):
            raise AcceleratorDefinitionError(
                "The following modifier files do not exist: "
                f"{', '.join([str(accel.model_dir / modifier) for modifier in inexistent_modifiers])}"
            )


class LhcBestKnowledgeCreator(LhcModelCreator):
    EXTRACTED_MQTS_FILENAME: str = "extracted_mqts.str"
    CORRECTIONS_FILENAME: str = "corrections.madx"

    @classmethod
    def get_madx_script(cls, accel: Lhc) -> str:
        if accel.excitation is not AccExcitationMode.FREE:
            raise AcceleratorDefinitionError("Don't set ACD or ADT for best knowledge model.")
        if accel.energy is None:
            raise AcceleratorDefinitionError("Best knowledge model requires energy.")

        corrections_file = accel.model_dir / cls.CORRECTIONS_FILENAME  # existence is tested in madx
        mqts_file = accel.model_dir / cls.EXTRACTED_MQTS_FILENAME  # existence is tested in madx

        madx_script = accel.get_base_madx_script(best_knowledge=True)
        madx_script += (
            f"call, file = '{corrections_file}';\n"
            f"call, file = '{mqts_file}';\n"
            f"exec, do_twiss_monitors(LHCB{accel.beam}, '{accel.model_dir / TWISS_BEST_KNOWLEDGE_DAT}', {accel.dpp});\n"
            f"exec, do_twiss_elements(LHCB{accel.beam}, '{accel.model_dir / TWISS_ELEMENTS_BEST_KNOWLEDGE_DAT}', {accel.dpp});\n"
        )
        return madx_script

    @classmethod
    def check_run_output(cls, accel: Lhc) -> None:
        files_to_check = [TWISS_BEST_KNOWLEDGE_DAT, TWISS_ELEMENTS_BEST_KNOWLEDGE_DAT]
        cls._check_files_exist(accel.model_dir, files_to_check)


class LhcCouplingCreator(LhcModelCreator):
    @classmethod
    def get_madx_script(cls, accel: Lhc) -> str:
        return cls.get_correction_check_script(accel)



"""
Plot Optics Measurements
------------------------

Wrapper for `plot_tfs` to easily plot the results from optics measurements.

.. code-block:: python

    from omc3.plotting.plot_optics_measurements import plot

    figs = plot(
        folders=['folder1', 'folder2'],
        combine_by=['files'],  # to compare folder1 and folder2
        output='output_directory',
        delta=True,  # delta from reference
        optics_parameters=['orbit', 'beta_phase', 'beta_amplitude',
                           'phase', 'total_phase',
                           'f1001_x', 'f1010_x'],
        x_axis='location',  # or 'phase-advance'
        ip_positions='LHCB1',
        suppress_column_legend=True,
        show=True,
        ncol_legend=2,
    )

**Arguments:**

*--Required--*

- **folders** *(MultiClass)*: Optics Measurements folders containing the analysed data.

- **optics_parameters** *(str)*: Optics parameters to plot, e.g. 'beta_amplitude'.
  RDTs need to be specified with plane, e.g. 'f1001_x'


*--Optional--*

- **change_marker**: Changes marker for each line in the plot.

  Action: ``store_true``
- **combine_by**: Combine plots into one. Either files, planes (not separated into two axes) or both.

  Choices: ``['files', 'planes']``
- **delta**: Plot the difference to model instead of the parameter.

  Action: ``store_true``
- **errorbar_alpha** *(float)*: Alpha value for error bars

  Default: ``0.6``
- **ip_positions**: Input to plot IP-Positions into the plots. Either 'LHCB1'
  or 'LHCB2' for LHC defaults, a dictionary of labels and positions or path to TFS file of a model.

- **ip_search_pattern**: In case your IPs have a weird name. Specify regex pattern.

  Default: ``IP\d$``
- **lines_manual** *(DictAsString)*: List of manual lines to plot.
  Need to contain arguments for axvline, and may contain the additional keys "text"
  and "loc" which is one of ['bottom', 'top', 'line bottom', 'line top']
  and places the text at the given location.

  Default: ``[]``
- **manual_style** *(DictAsString)*: Additional style rcParameters which
  update the set of predefined ones.

  Default: ``{}``
- **ncol_legend** *(int)*: Number of bpm legend-columns. If < 1 no legend is shown.

  Default: ``3``
- **output** *(MultiClass)*: Folder to output the results to.

- **plot_styles** *(str)*: Which plotting styles to use,
  either from plotting.styles.*.mplstyles or default mpl.

  Default: ``['standard']``
- **share_xaxis**: In case of multiple axes per figure, share x-axis.

  Action: ``store_true``
- **show**: Shows plots.

  Action: ``store_true``
- **suppress_column_legend**: Does not show column name in legend
  e.g. when combining by files (see also `ncol_legend`).

  Action: ``store_true``
- **x_axis**: Which parameter to use for the x axis.

  Choices: ``['location', 'phase-advance']``
  Default: ``location``
- **x_lim** *(float, int, None)*: Limits on the x axis (Tupel)

- **y_lim** *(float, int, None)*: Limits on the y axis (Tupel)
"""
from pathlib import Path

import tfs
from generic_parser import EntryPointParameters, entrypoint
from generic_parser.entry_datatypes import DictAsString

from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import ERR, DELTA, AMPLITUDE, PHASE, EXT
from omc3.optics_measurements.rdt import _rdt_to_order_and_type
from omc3.plotting.optics_measurements.constants import (DEFAULTS,
                                                         XAXIS, YAXIS,
                                                         IP_POS_DEFAULT)
from omc3.plotting.plot_tfs import plot as plot_tfs, float_or_none
from omc3.plotting.spectrum.utils import get_unique_filenames
from omc3.plotting.utils.lines import VERTICAL_LINES_TEXT_LOCATIONS
from omc3.utils.iotools import PathOrStr, save_config
from omc3.utils.logging_tools import get_logger, list2str

LOG = get_logger(__name__)


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="folders",
        help="Optics Measurements folders containing the analysed data.",
        required=True,
        nargs="+",
        type=PathOrStr,
    )
    params.add_parameter(
        name="optics_parameters",
        help=("Optics parameters to plot, e.g. 'beta_amplitude'. "
              "RDTs need to be specified with plane, e.g. 'f1001_x'"),
        required=True,
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="delta",
        help="Plot the difference to model instead of the parameter.",
        action="store_true"
    )
    params.add_parameter(
        name="ip_positions",
        help=("Input to plot IP-Positions into the plots. "
              "Either 'LHCB1' or 'LHCB2' for LHC defaults, "
              "a dictionary of labels and positions "
              "or path to TFS file of a model."),
    )
    params.add_parameter(name="lines_manual",
                         nargs="*",
                         default=[],
                         type=DictAsString,
                         help='List of manual lines to plot. Need to contain arguments for axvline, and may contain '
                              'the additional keys "text" and "loc" which is one of '
                              f'{list(VERTICAL_LINES_TEXT_LOCATIONS.keys())} and places the text at the given location.'
                         )
    params.add_parameter(
        name="ip_search_pattern",
        default=r"IP\d$",
        help="In case your IPs have a weird name. Specify regex pattern.",
    )
    params.add_parameter(
        name="x_axis",
        help="Which parameter to use for the x axis.",
        choices=list(XAXIS.keys()),
        default='location',
    )
    # Parameters that are only passed on ---
    params.add_parameter(
        name="combine_by",
        help="Combine plots into one. Either files, planes (not separated into two axes) or both.",
        nargs="+",
        choices=['files', 'planes'],  # combine by columns does not really make sense
    )
    params.add_parameter(
        name="output",
        help="Folder to output the results to.",
        type=PathOrStr,
    )
    params.add_parameter(
        name="show",
        help="Shows plots.",
        action="store_true",
    )
    params.add_parameter(
        name="plot_styles",
        type=str,
        nargs="+",
        default=['standard'],
        help='Which plotting styles to use, either from plotting.styles.*.mplstyles or default mpl.'
    )
    params.add_parameter(
        name="manual_style",
        type=DictAsString,
        default={},
        help='Additional style rcParameters which update the set of predefined ones.'
    )
    params.add_parameter(
        name="change_marker",
        help="Changes marker for each line in the plot.",
        action="store_true",
    )
    params.add_parameter(
        name="ncol_legend",
        type=int,
        default=DEFAULTS['ncol_legend'],
        help='Number of bpm legend-columns. If < 1 no legend is shown.'
    )
    params.add_parameter(
        name="suppress_column_legend",
        help=("Does not show column name in legend "
              "e.g. when combining by files (see also `ncol_legend`)."),
        action="store_true",
    )
    params.add_parameter(
        name="errorbar_alpha",
        help="Alpha value for error bars",
        type=float,
        default=DEFAULTS['errorbar_alpha'],
    )
    params.add_parameter(
        name="x_lim",
        nargs=2,
        type=float_or_none,
        help='Limits on the x axis (Tupel)'
    )
    params.add_parameter(
        name="y_lim",
        nargs=2,
        type=float_or_none,
        help='Limits on the y axis (Tupel)'
    )
    params.add_parameter(
        name="share_xaxis",
        help="In case of multiple axes per figure, share x-axis.",
        action="store_true",
    )
    return params


@entrypoint(get_params(), strict=True)
def plot(opt):
    """Main plotting function."""
    LOG.info("Starting plotting of optics measurement data: "
             f"{list2str(opt.optics_parameters)} in {list2str(opt.folders):s}")

    if opt.output is not None:
        save_config(Path(opt.output), opt, __file__)

    opt = _check_opt(opt)

    ip_positions = _get_ip_positions(opt.ip_positions, opt.x_axis, opt.ip_search_pattern)
    x_column, x_label = _get_x_options(opt.x_axis)

    fig_dict = {}
    for optics_parameter in opt.optics_parameters:

        is_rdt = optics_parameter.lower().startswith("f")
        files, file_labels = zip(*get_unique_filenames(opt.folders))
        file_labels = ["_".join(flabels) for flabels in file_labels]

        if is_rdt:
            fig_dict.update(_plot_rdt(
                optics_parameter, files, file_labels, x_column, x_label,
                ip_positions, opt,)
            )
        else:
            if not optics_parameter.endswith("_"):
                optics_parameter += "_"

            fig_dict.update(_plot_param(
                optics_parameter, files, file_labels, x_column, x_label,
                ip_positions, opt,)
            )
    return fig_dict


def _check_opt(opt):
    return opt


# Plot RDTs --------------------------------------------------------------------


def _plot_rdt(optics_parameter, files, file_labels, x_column, x_label, ip_positions, opt):
    """Main plotting function for RDTs."""
    fig_dict = {}
    if opt.x_axis != 'location':
        LOG.error("Phase advance not yet implemented in RDT-files. Skipping.")
        return fig_dict

    if opt.delta:
        LOG.warning('Delta Columns for RDTs not implemented. Using normal columns.')

    subfolder = _rdt_to_order_and_type([int(n) for n in optics_parameter[1:5]])
    files = [str(f.absolute()/'rdt'/subfolder/f'{optics_parameter}{EXT}') for f in files]
    columns = _get_rdt_columns()

    combine_by = []
    if opt.combine_by is not None:
        combine_by = list(opt.combine_by)

    prefix = ''
    if "files" in combine_by:
        prefix = f'{optics_parameter}_'

    combine_planes = False
    if "planes" in combine_by:
        combine_by[combine_by.index("planes")] = "columns"
        combine_planes = True

    for idxs, mode in ((slice(0, 2), "amplitude"), (slice(2, 4), "complex")):
        if combine_planes:
            y_labels = [optics_parameter.upper()]
            if opt.suppress_column_legend:
                column_labels = [l.format('F') for l in columns['y_labels'][idxs]]
            else:
                column_labels = [l.format(optics_parameter.upper()) for l in columns['y_labels'][idxs]]

        else:
            y_labels = [l.format(optics_parameter.upper()) for l in columns['y_labels'][idxs]]
            column_labels = [optics_parameter.upper()]
            if opt.suppress_column_legend:
                column_labels = ['']

        fig_dict.update(
            plot_tfs(
                files=files,
                file_labels=list(file_labels),
                y_columns=columns['y_columns'][idxs],
                column_labels=column_labels,
                y_labels=[y_labels],
                error_columns=columns['error_columns'][idxs],
                x_columns=[x_column],
                x_labels=[x_label],
                vertical_lines=ip_positions + opt.lines_manual,
                same_figure="columns" if "columns" not in combine_by else None,
                same_axes=combine_by if len(combine_by) else None,
                single_legend=True,
                output_prefix=f"plot_{mode}_{prefix}",
                **opt.get_subdict(['show', 'output',
                                   'plot_styles', 'manual_style',
                                   'change_marker', 'errorbar_alpha',
                                   'ncol_legend', 'x_lim', 'y_lim',
                                   'share_xaxis'])
            ))

    return fig_dict


def _get_rdt_columns():
    rdt_measures = ['rdt_amp', 'rdt_phase', 'rdt_real', 'rdt_imag']
    result = {key: [None] * len(rdt_measures) for key in ['y_columns', 'y_labels', 'error_columns']}
    for idx, meas in enumerate(rdt_measures):
        column, _, label = YAXIS[meas]
        result['y_columns'][idx] = column
        result['error_columns'][idx] = f"{ERR}{AMPLITUDE}"  # replaced for phase below
        result['y_labels'][idx] = label
    result['error_columns'][rdt_measures.index('rdt_phase')] = f"{ERR}{PHASE}"
    return result


# Plot Other Parameter ---------------------------------------------------------


def _plot_param(optics_parameter, files, file_labels, x_column, x_label, ip_positions, opt):
    """Main plotting function for all parameters but RDTs."""
    y_column, error_column, column_label, y_label = _get_columns_and_label(optics_parameter, opt.delta)

    same_fig = None
    column_labels = [column_label]
    if opt.combine_by is None or 'planes' not in opt.combine_by:
        same_fig = 'planes'

    if opt.combine_by is not None and "planes" in opt.combine_by:  # not else!
        column_labels = [f'{column_label} {{0}}']

    if opt.suppress_column_legend:
        if same_fig == 'planes':
            column_labels = ['']
        else:
            column_labels = ['{0}']   #  show planes in labels as all are in same axes

    prefix = ''
    if opt.delta:
        prefix += f'delta_'
    if opt.combine_by and "files" in opt.combine_by:
        prefix += f'{optics_parameter}'

    return plot_tfs(
        files=[f.absolute()/f'{optics_parameter}{{0}}{EXT}' for f in files],
        file_labels=list(file_labels),
        y_columns=[f'{y_column}{{0}}'],
        y_labels=[[y_label]],
        column_labels=column_labels,
        error_columns=[f'{error_column}{{0}}'],
        x_columns=[x_column],
        x_labels=[x_label],
        planes=list(PLANES),
        vertical_lines=ip_positions + opt.lines_manual,
        same_figure=same_fig,
        same_axes=opt.combine_by,
        single_legend=True,
        output_prefix=f"plot_{prefix}",
        **opt.get_subdict(['show', 'output',
                           'plot_styles', 'manual_style',
                           'change_marker', 'errorbar_alpha',
                           'ncol_legend', 'x_lim', 'y_lim',
                           'share_xaxis'])
    )


def _get_columns_and_label(parameter, delta):
    column_and_label = YAXIS[parameter]
    column = column_and_label[0]
    column_label = column_and_label[1]
    y_label = column_and_label[2]
    if delta:
        column = f"{DELTA}{column}"
        try:
            y_label = column_and_label[3]
        except IndexError:
            y_label = fr'$\Delta {y_label[1:]}'

    return column, f"{ERR}{column}", column_label, y_label


# IP-Positions -----------------------------------------------------------------


def _get_ip_positions(ip_positions, xaxis, ip_pattern):
    if isinstance(ip_positions, (str, Path)):
        try:
            positions = IP_POS_DEFAULT[ip_positions]
        except KeyError:
            return _get_ip_positions_from_file(ip_positions, xaxis, ip_pattern)
        else:
            if xaxis != 'location':
                raise NotImplementedError("No default IP positions for "
                                          f"{xaxis} implemented.")
        return _create_ip_list(positions)

    if ip_positions is None:
        return []

    return ip_positions


def _get_ip_positions_from_file(path, axis, pattern):
    model = tfs.read(path, index="NAME")
    ip_mask = model.index.str.match(pattern)
    column = XAXIS[axis][0]
    return _create_ip_list(model.loc[ip_mask, column])


def _create_ip_list(ip_dict):
    """ip_dict can be series as well."""
    return [{'text': name, 'x': x,
             'loc': 'top', 'color': 'k'} for name, x in ip_dict.items()]


# Other ------------------------------------------------------------------------


def _get_x_options(x_axis):
    return XAXIS[x_axis]


# Script Mode ------------------------------------------------------------------


if __name__ == '__main__':
    plot()



"""
MAD-X Wrapper
-------------

``madx_wrapper`` is high-level wrapper for running ``MAD-X`` codes within ``omc3``.

Usage: ``python madx_wrapper.py --file your_madx_file.madx``
"""
import contextlib
import os
import subprocess
import sys
import warnings
from pathlib import Path
from tempfile import mkstemp

from generic_parser import EntryPointParameters, entrypoint

from omc3.utils import logging_tools

LOG = logging_tools.get_logger(__name__)

_LOCAL_PATH = Path(__file__).parent / "bin"

if "darwin" in sys.platform:
    _MADX_BIN = "madx-macosx64-gnu"
elif "win" in sys.platform:
    _MADX_BIN = "madx-win64-gnu.exe"
else:
    _MADX_BIN = "madx-linux64-gnu"

MADX_PATH = _LOCAL_PATH / _MADX_BIN
warnings.simplefilter('always', DeprecationWarning)


class MadxError(Exception):
    pass


def madx_wrapper_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="file",
        required=True,
        help="The file with the annotated MADX input to run."
    )
    params.add_parameter(
        name="output",
        help="Path to a file where to write the MADX script."
    )
    params.add_parameter(
        name="log",
        help="Path to a file where to write the MADX log output."
    )
    params.add_parameter(
        name="madx_path",
        default=MADX_PATH,
        help="Path to the MAD-X executable to use"
    )
    params.add_parameter(
        name="cwd",
        help="Set current working directory"
    )
    return params


@entrypoint(madx_wrapper_params(), strict=True)
def main(opt):
    run_file(
        input_file=Path(opt.file),
        output_file=Path(opt.output) if opt.output else None,
        log_file=Path(opt.log) if opt.log else None,
        madx_path=Path(opt.madx_path),
        cwd=Path(opt.cwd) if opt.cwd else None
    )


def run_file(input_file: Path, output_file: Path = None, log_file: Path = None,
             madx_path: Path = MADX_PATH, cwd: Path = None):
    """
    Runs ``MAD-X`` on a given script in a subprocess.

    Args:
        input_file: ``MAD-X`` input file.
        output_file: If given writes ``MAD-X`` script.
        log_file: If given writes ``MAD-X`` logging output.
        madx_path: Path to the ``MAD-X`` executable.
    """
    input_string = input_file.read_text()
    run_string(input_string, output_file=output_file, log_file=log_file,
               madx_path=madx_path, cwd=cwd)


def run_string(input_string: str,
               output_file: Path = None, log_file: Path = None,
               madx_path: Path = MADX_PATH, cwd: Path = None):
    """
    Runs ``MAD-X`` on a given string in a subprocess.

    Args:
        input_string: ``MAD-X`` input string.
        output_file: If given writes ``MAD-X`` script.
        log_file: If given writes ``MAD-X`` logging output.
        madx_path: Path to the ``MAD-X`` executable.
    """
    _check_log_and_output_files(output_file, log_file)
    _run(input_string, log_file, output_file, madx_path, cwd)


def _run(full_madx_script, log_file=None, output_file=None, madx_path=MADX_PATH, cwd=None):
    """Starts the ``MAD-X`` process."""
    with _madx_input_wrapper(full_madx_script, output_file) as madx_jobfile:
        process = subprocess.Popen(
            [str(Path(madx_path).absolute()), str(Path(madx_jobfile).absolute())],
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=None if cwd is None else str(cwd),
        )
        with _logfile_wrapper(log_file) as log_handler, process.stdout:
            for line in process.stdout:
                log_handler(line.decode('utf-8'))
        status = process.wait()

    if status:
        _raise_madx_error(log=log_file, file=output_file)

# Wrapper ####################################################################


def _check_log_and_output_files(output_file, log_file):
    """Check read/write access. The EAFP way."""
    if output_file is not None:
        open(output_file, "a").close()
    if log_file is not None:
        open(log_file, "a").close()


@contextlib.contextmanager
def _logfile_wrapper(file_path=None):
    """Logs into file and debug if file is given, into info otherwise."""
    if file_path is None:
        def log_handler(line):
            line = line.rstrip()
            if len(line):
                LOG.log(logging_tools.MADX, line)
        yield log_handler
    else:
        with open(file_path, "w") as log_file:
            def log_handler(line):
                log_file.write(line)
                line = line.rstrip()
                if len(line):
                    LOG.log(logging_tools.MADX, line)
            yield log_handler


@contextlib.contextmanager
def _madx_input_wrapper(content, file_path=None):
    """
    Writes content into an output file and returns filepath.
    If ``file_path`` is not given, the output file is temporary and will be deleted afterwards.
    """
    if file_path is None:
        temp_file = True
        fd, file_path = mkstemp(suffix=".madx", prefix="job.", text=True)
        os.close(fd)  # close file descriptor
        if content:
            with open(file_path, "w") as f:
                f.write(content)
    else:
        temp_file = False
        with open(file_path, "w") as f:
            f.write(content)
    try:
        yield file_path
    finally:
        if temp_file:
            os.remove(file_path)


def _raise_madx_error(log=None, file=None):
    """
    Raising Error Wrapper. Extracts extra info from log and output file if given.
    """
    message = "MADX run failed."
    if log is not None:
        try:
            with open(log, "r") as f:
                content = f.readlines()
            if content[-1].startswith("+="):
                message += f" '{content[-1].replace('+=+=+=', '').strip()}'."
        except (IOError, IndexError):
            pass

    if file is not None:
        message += f" Run on File: '{file}'."

    raise MadxError(message)


if __name__ == "__main__":
    main()



"""
Plotting Utilities: Annotations
-------------------------------

Helper functions to create annotations, legends as well as style labels in plots.
"""
import itertools
import re
from typing import Union

import matplotlib
import pandas as pd
import tfs
from matplotlib import pyplot as plt

# List of common y-labels. Sorry for the ugly.
ylabels = {
    "beta":               r'$\beta_{{{0}}} \quad [m]$',
    "betabeat":           r'$\Delta \beta_{{{0}}} \; / \; \beta_{{{0}}}$',
    "betabeat_permile":   r'$\Delta \beta_{{{0}}} \; / \; \beta_{{{0}}} [$'u'\u2030'r'$]$',
    "dbeta":              r"$\beta'_{{{0}}} \quad [m]$",
    "dbetabeat":          r'$1 \; / \; \beta_{{{0}}} \cdot \partial\beta_{{{0}}} \; / \; \partial\delta_{{{0}}}$',
    "norm_dispersion":    r'D$_{{{0}}} \; / \; \sqrt{{\beta_{{{0}}}}} \quad \left[\sqrt{{\rm m}}\right]$',
    "norm_dispersion_mu": r'D$_{{{0}}} \; / \; \sqrt{{\beta_{{{0}}}}} \quad \left[\sqrt{{\rm \mu m}}\right]$',
    "phase":              r'$\phi_{{{0}}} \quad [2\pi]$',
    "phasetot":           r'$\phi_{{{0}}} \quad [2\pi]$',
    "phase_milli":        r'$\phi_{{{0}}} \quad [2\pi\cdot10^{{-3}}]$',
    "dispersion":         r'D$_{{{0}}}$ [m]',
    "dispersion_mm":      r'D$_{{{0}}}$ [mm]',
    "co":                 r'Orbit {0} [m]',
    "co_mm":              r'Orbit {0} [mm]',
    "tune":               r'Q$_{{{0}}}$',
    "nattune":            r'Nat. Q$_{{{0}}}$',
    "chromamp":           r'W$_{{{0}}}$',
    "real":               r'$\Re({0})$',
    "imag":               r'$\Im({0})$',
    "absolute":           r'$\left|{0}\right|$',
}


def set_yaxis_label(param, plane, ax=None, delta=False, chromcoup=False):  # plot x and plot y
    """
    Set y-axis labels.

    Args:
        param: One of the labels defined in ``ylabel`` in this module.
        plane: Usually x or y, but can be any string actually to be placed into the label ({0}).
        ax: Axes to put the label on. Defaults to `` gca()``.
        delta: If True adds a Delta before the label. Defaults to ``False``.
    """
    if not ax:
        ax = plt.gca()
    try:
        label = ylabels[param].format(plane)
    except KeyError:
        raise ValueError(f"Label '{param}' not found.")

    if delta:
        if param.startswith("beta") or param.startswith("norm"):
            label = fr'$\Delta({label[1:-1]})$'
        else:
            label = fr'$\Delta {label[1:]}'

    if chromcoup:
        label = fr'{label[:-1] }/\Delta\delta$'

    ax.set_ylabel(label)


def set_xaxis_label(ax=None):
    """
    Sets the standard x-axis label

    Args:
        ax: Axes to put the label on. Defaults to ``gca()``.
    """
    if not ax:
        ax = plt.gca()
    ax.set_xlabel(r'Longitudinal location [m]')


def show_ir(ip_dict, ax=None, mode='inside'):
    """
    Plots the interaction regions into the background of the plot.

    Args:
        ip_dict: `dict`, `dataframe` or `series` containing ``IPLABEL`` : ``IP_POSITION``.
        ax:  Axes to put the irs on. Defaults to gca()``.
        mode: ``inside``, ``outside`` + ``nolines`` or just ``lines``.
    """
    if ax is None:
        ax = plt.gca()

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    lines = 'nolines' not in mode
    inside = 'inside' in mode
    lines_only = 'inside' not in mode and 'outside' not in mode and 'lines' in mode

    if isinstance(ip_dict, (pd.DataFrame, pd.Series)):
        if isinstance(ip_dict, pd.DataFrame):
            ip_dict = ip_dict.iloc[:, 0]
        d = {}
        for ip in ip_dict.index:
            d[ip] = ip_dict.loc[ip]
        ip_dict = d

    for ip in ip_dict.keys():
        if xlim[0] <= ip_dict[ip] <= xlim[1]:
            xpos = ip_dict[ip]

            if lines:
                ax.axvline(xpos, linestyle=':', color='grey', marker='', zorder=0)

            if not lines_only:
                ypos = ylim[not inside] + (ylim[1] - ylim[0]) * 0.01
                c = 'grey' if inside else matplotlib.rcParams["text.color"]
                ax.text(xpos, ypos, ip, color=c, ha='center', va='bottom')

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)


def move_ip_labels(value, ax=None):
    """Moves IP labels according to max y * value."""
    if ax is None:
        ax = plt.gca()

    y_max = ax.get_ylim()[1]
    for t in ax.texts:
        if re.match(r"^IP\s*\d$", t.get_text()):
            x = t.get_position()[0]
            t.set_position((x, y_max * value))


def get_ip_positions(path):
    """
    Returns a `dict` of IP positions from tfs-file of path.

    Args:
        path (str): `Path` to the tfs-file containing IP-positions.
    """
    df = tfs.read_tfs(path).set_index('NAME')
    ip_names = [f"IP{i:d}" for i in range(1, 9)]
    ip_pos = df.loc[ip_names, 'S'].to_numpy()
    return dict(zip(ip_names, ip_pos))


def set_name(name, fig_or_ax=None):
    """
    Sets the name of the figure or axes.

    Args:
        name (str): string to set as name.
        fig_or_ax: `Figure` or `Axes` to to use. If ``None`` is given, takes the current figure.
            Defaults to ``None``.
    """
    if not fig_or_ax:
        fig_or_ax = plt.gcf()
    try:
        fig_or_ax.figure.canvas.manager.set_window_title(name)
    except AttributeError:
        fig_or_ax.canvas.manager.set_window_title(name)


def get_name(fig_or_ax=None):
    """
    Returns the name of the figure or axes.

    Args:
        fig_or_ax: `Figure` or `Axes` to to use.If ``None`` is given, takes the current figure.
            Defaults to ``None``.
    """
    if not fig_or_ax:
        fig_or_ax = plt.gcf()
    try:
        return fig_or_ax.figure.canvas.get_window_title()
    except AttributeError:
        return fig_or_ax.canvas.get_window_title()


def set_annotation(text, ax=None, position='right', pad=0.02):
    """
    Writes an annotation on the top right of the axes.

    Args:
        text: The annotation.
        ax: `Axes` to set annotation on. If ``None`` is given, takes the current figure. Defaults
            to ``None``.
        position: ``left`` or ``right``.
        pad: padding to the axes.
    """
    if ax is None:
        ax = plt.gca()

    annotation = get_annotation(ax, by_reference=True)

    xpos = float(position == 'right')

    if annotation is None:
        ax.text(xpos, 1.0 + pad, text,
                ha=position,
                va='bottom',
                transform=ax.transAxes,
                label='plot_style_annotation')
    else:
        annotation.set_text(text)
        plt.setp(annotation, ha=position, x=xpos, y=1.0+pad)


def get_annotation(ax=None, by_reference=True):
    """
    Returns the annotation set by ``set_annotation()``.

    Args:
        ax: `Axes` to get annotation from. If ``None`` is given, takes the current figure.
            Defaults to ``None``.
        by_reference (bool): If ``True``, returns the reference to the annotation, otherwise the
            text as string. Defaults to ``True``.
    """
    if not ax:
        ax = plt.gca()

    for c in ax.get_children():
        if c.get_label() == 'plot_style_annotation':
            if by_reference:
                return c
            else:
                return c.get_text()
    return None


def small_title(ax=None):
    """
    Alternative to annotation, which lets you use the title-functions.

    Args:
        ax: `Axes` to use. If ``None`` is given, takes the current axes. Defaults to ``None``.
    """
    if not ax:
        ax = plt.gca()

    # could not get set_title() to work properly, so one parameter at a time
    ax.title.set_position([1.0, 1.02])
    ax.title.set_transform(ax.transAxes)
    ax.title.set_fontsize(matplotlib.rcParams['font.size'])
    ax.title.set_fontweight(matplotlib.rcParams['font.weight'])
    ax.title.set_verticalalignment('bottom')
    ax.title.set_horizontalalignment('right')


def figure_title(text, ax=None, pad=0, **kwargs):
    """
    Set the title all the way to the top.

    Args:
        text: Text for the title.
        ax: `Axes` to use. If ``None`` is given, takes the current axes. Defaults to ``None``.
        pad: Padding from border.
        kwargs: passed on to fontdict.
    """
    if not ax:
        ax = plt.gca()

    # could not get set_title() to work properly, so one parameter at a time
    fdict = dict(fontsize=matplotlib.rcParams['font.size'],
                 fontweight=matplotlib.rcParams['font.weight'],
                 va="top", ha="center")
    fdict.update(kwargs)
    ax.set_title(text, transform=ax.figure.transFigure, fontdict=fdict)

    ax.title.set_position([.5, float(fdict['va'] == "top") + pad * (-1)**(fdict['va'] == 'top')])


def get_legend_ncols(labels, max_length=78):
    """Calculate the number of columns in legend dynamically."""
    return max([max_length/max([len(l) for l in labels]), 1])


def transpose_legend_order(ncol, handles=None, labels=None, ax=None):
    """ Reorder handles and labels, so that the legend order is transposed,
    i.e. the entries are row-first instead of column-first."""
    def reorder(items):
        return list(itertools.chain(*[items[i::ncol] for i in range(ncol)]))

    if handles is None or labels is None:
        if ax is None:
            ax = plt.gca()

        ax_handles, ax_labels = ax.get_legend_handles_labels()
        if handles is None:
            handles = ax_handles

        if labels is None:
            labels = ax_labels

    return reorder(handles), reorder(labels)


def make_top_legend(ax, ncol, frame=False, handles=None, labels=None, pad=0.02, transposed=False):
    """Create a legend on top of the plot."""
    if ncol < 1:
        return

    if transposed:
        handles, labels = transpose_legend_order(ncol, handles, labels, ax)

    leg = ax.legend(handles=handles, labels=labels, loc='lower right',
                    bbox_to_anchor=(1.0, 1.0+pad),
                    fancybox=frame, shadow=frame, frameon=frame, ncol=ncol)

    leg.axes.figure.canvas.draw()
    legend_width = leg.get_window_extent().transformed(leg.axes.transAxes.inverted()).width
    if legend_width > 1:
        x_shift = (legend_width - 1) / 2.
        ax.legend(handles=handles, labels=labels, loc='lower right',
                  bbox_to_anchor=(1.0 + x_shift, 1.0+pad),
                  fancybox=frame, shadow=frame, frameon=frame, ncol=ncol)

    return leg


class OOMFormatter(matplotlib.ticker.ScalarFormatter):
    """
    Order of Magnitude Formatter.

    To set a fixed order of magnitude and fixed significant numbers.
    As seen on: https://stackoverflow.com/a/42658124/5609590

    See: set_sci_magnitude
    """
    def __init__(self, order=0, fformat="%1.1f", offset=True, mathText=True):
        self.oom = order
        self.fformat = fformat
        matplotlib.ticker.ScalarFormatter.__init__(self, useOffset=offset, useMathText=mathText)

    def _set_orderOfMagnitude(self, nothing):
        self.orderOfMagnitude = self.oom

    def _set_format(self, vmin, vmax):
        self.format = self.fformat
        if self._useMathText:
            self.format = '$%s$' % matplotlib.ticker._mathdefault(self.format)


def set_sci_magnitude(ax, axis="both", order=0, fformat="%1.1f", offset=True, math_text=True):
    """
    Uses the OMMFormatter to set the scientific limits on axes.

    Args:
        ax: Plotting axes.
        axis (str): **x**, **y** or **both**.
        order (int): Magnitude Order.
        fformat (str): Format to use.
        offset (bool): Formatter offset.
        math_text (bool): Whether to use mathText.
    """
    oomf = OOMFormatter(order=order, fformat=fformat, offset=offset, mathText=math_text)

    if axis == "x" or axis == "both":
        ax.xaxis.set_major_formatter(oomf)

    if axis == "y" or axis == "both":
        ax.yaxis.set_major_formatter(oomf)

    ax.ticklabel_format(axis=axis, style="sci", scilimits=(order, order), useMathText=math_text)

    return ax


def get_fontsize_as_float(font_size: Union[str, float]) -> float:
    try:
        scale = {
            'xx-small': 0.579,
            'x-small': 0.694,
            'small': 0.833,
            'medium': 1.0,
            'large': 1.200,
            'x-large': 1.440,
            'xx-large': 1.728,
            'larger': 1.2,
            'smaller': 0.833,
            None: 1.0}[font_size]
    except KeyError:
        return font_size
    return scale * matplotlib.rcParams['font.size']



"""
BetaBeat.src Output Converter
-------------

Script to convert most important output files produced by ``BetaBeat.src`` / ``GetLLM`` into the standard
format used in ``omc3`` to allow straight forward comparison of the two.

Good to know: in ``BetaBeat.src`` the ``_free`` files correspond to AC dipole (or ADT-AC dipole)
compensation by analytic equation while ``_free2`` corresponds to compensation via effective model. It is
written in GetLLM (at least once) here: https://github.com/pylhc/Beta-Beat.src/blob/63c5e39f63b03c00d18289cc9813a912fa6b933f/GetLLM/GetLLM.py#L606
"""
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Union

import tfs
from generic_parser.entrypoint_parser import (
    EntryPointParameters,
    entrypoint,
    save_options_to_config,
    DotDict,
)

from omc3.definitions import formats
from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import (AMPLITUDE, AMP_BETA_NAME,BETA_NAME, DELTA, DISPERSION_NAME,
                                                ERR, EXT, IMAG, MDL, NORM_DISP_NAME, ORBIT_NAME, PHASE,
                                                PHASE_NAME, REAL, TOTAL_PHASE_NAME)
from omc3.optics_measurements.toolbox import df_ang_diff, df_diff, df_err_sum, df_ratio, df_rel_diff
from omc3.utils import contexts, iotools, logging_tools

LOGGER = logging_tools.get_logger(__name__)
OLD_EXT: str = ".out"
DEFAULT_CONFIG_FILENAME: str = "old_measurement_converter_{time:s}.ini"


def converter_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="inputdir",
        required=True,
        type=str,
        help="Directory with BetaBeat.src output files."
    )
    params.add_parameter(
        name="outputdir", required=True, type=str, help="Output directory for converted files."
    )
    params.add_parameter(
        name="suffix",
        type=str,
        default="_free",
        choices=("", "_free", "_free2"),
        help="AC dipole compensation suffix used in the provided BetaBeat.src output ('_free' "
             "for compensation by equation, '_free2' by model. Defaults to '_free'.",
    )
    return params


@entrypoint(converter_params(), strict=True)
def converter_entrypoint(opt: DotDict) -> None:
    """
    Looks for expected ``BetaBeat.src`` output files in the provided input directory, converts them to
    the format used in ``omc3`` and writes the converted files in the provided output directory.

    *--Required--*

    - **inputdir** *(str)*:

        Directory with BetaBeat.src output files.


    - **outputdir** *(str)*:

        Output directory for converted files.


    *--Optional--*

    - **suffix** *(str)*:

        AC dipole compensation suffix used in the provided BetaBeat.src output ('_free' for
        compensation by equation, '_free2' by model).

        choices: ``('', '_free', '_free2')``

        default: ``_free``

    """
    iotools.create_dirs(Path(opt.outputdir))
    save_options_to_config(
        Path(opt.outputdir) / DEFAULT_CONFIG_FILENAME.format(time=datetime.utcnow().strftime(formats.TIME)),
        OrderedDict(sorted(opt.items())),
    )
    LOGGER.warning("Be aware that the input units in the BetaBeat.src outputs may not be SI")
    convert_old_directory_to_new(opt)


def convert_old_directory_to_new(opt: DotDict) -> None:
    """
    Looks in the provided directory for expected ``BetaBeat.src`` output files, converts it to the output
    format used by ``omc3`` and  write them to the new location.

    Args:
        opt (DotDict): The entrypoint parameters parsed from the command line.
    """
    inputdir_path = Path(opt.inputdir)
    outputdir_path = Path(opt.outputdir)
    suffix = opt.suffix
    LOGGER.info(f"Converting old BetaBeat.src outputs at '{inputdir_path.absolute()}' to omc3 format")

    with contexts.timeit(lambda spanned: LOGGER.info(f"Total time for conversion: {spanned}s")):
        for plane in PLANES:
            convert_old_beta_from_amplitude(inputdir_path, outputdir_path, suffix, plane)
            convert_old_beta_from_phase(inputdir_path, outputdir_path, suffix, plane)
            convert_old_phase(inputdir_path, outputdir_path, suffix, plane)
            convert_old_total_phase(inputdir_path, outputdir_path, suffix, plane)
            # TODO phase vs phasetot inconsistent naming NAME S , first and second BPMs swapped locations
            convert_old_closed_orbit(inputdir_path, outputdir_path, plane)
            convert_old_dispersion(inputdir_path, outputdir_path, plane)
        convert_old_coupling(inputdir_path, outputdir_path, suffix)
        convert_old_normalised_dispersion(inputdir_path, outputdir_path, "X")


def convert_old_beta_from_amplitude(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    suffix: str,
    plane: str,
    old_file_name: str = "ampbeta",
    new_file_name: str = AMP_BETA_NAME,
) -> None:
    """
    Looks in the provided directory for expected beta from amplitude file from ``BetaBeat.src`` for a given
    plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getampbeta(x,y).out, with the following expected columns: NAME, S, COUNT,
    BETX, BETXSTD, BETXMDL, MUXMDL, BETXRES, BETXSTDRES.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        suffix (str): Compensation suffix used in the provided BetaBeat.src output files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old beta from amplitude file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{suffix}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    dframe = dframe.rename(
        columns={f"BET{plane}STD": f"{ERR}BET{plane}", f"BET{plane}STDRES": f"{ERR}BET{plane}RES"},
    )
    dframe[f"{DELTA}BET{plane}"] = df_rel_diff(dframe, f"BET{plane}", f"BET{plane}{MDL}")
    dframe[f"{ERR}{DELTA}BET{plane}"] = df_ratio(dframe, f"{ERR}BET{plane}", f"BET{plane}{MDL}")
    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_beta_from_phase(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    suffix: str,
    plane: str,
    old_file_name: str = "beta",
    new_file_name: str = BETA_NAME,
) -> None:
    """
    Looks in the provided directory for expected beta from phase file from ``BetaBeat.src`` for a given
    plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getbeta(x,y).out, with the following expected columns: NAME, S, COUNT,
    BETX, SYSBETX, STATBETX, ERRBETX, CORR_ALFABETA, ALFX, SYSALFX, STATALFX, ERRALFX, BETXMDL, ALFXMDL,
    MUXMDL, NCOMBINATIONS.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        suffix (str): Compensation suffix used in the provided BetaBeat.src output files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old beta from phase file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{suffix}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    if "CORR_ALFABETA" in dframe.columns.to_numpy():
        dframe = dframe.drop(
            columns=[f"STATBET{plane}", f"SYSBET{plane}", "CORR_ALFABETA", f"STATALF{plane}", f"SYSALF{plane}"]
        )
    else:
        dframe[f"{ERR}BET{plane}"] = df_err_sum(dframe, f"{ERR}BET{plane}", f"STDBET{plane}")
        dframe[f"{ERR}ALF{plane}"] = df_err_sum(dframe, f"{ERR}ALF{plane}", f"STDALF{plane}")

    dframe[f"{DELTA}BET{plane}"] = df_rel_diff(dframe, f"BET{plane}", f"BET{plane}{MDL}")
    dframe[f"{ERR}{DELTA}BET{plane}"] = df_ratio(dframe, f"{ERR}BET{plane}", f"BET{plane}{MDL}")
    dframe[f"{DELTA}ALF{plane}"] = df_diff(dframe, f"ALF{plane}", f"ALF{plane}{MDL}")
    dframe[f"{ERR}{DELTA}ALF{plane}"] = dframe.loc[:, f"{ERR}ALF{plane}"].to_numpy()
    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_phase(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    suffix: str,
    plane: str,
    old_file_name: str = "phase",
    new_file_name: str = PHASE_NAME,
) -> None:
    """
    Looks in the provided directory for expected phase file from ``BetaBeat.src`` for a given
    plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getphase(x,y).out, with the following expected columns: NAME, NAME2, S, S1,
    COUNT, PHASEX, STDPHX, PHXMDL, MUXMDL.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        suffix (str): Compensation suffix used in the provided BetaBeat.src output files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old phase file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{suffix}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    dframe = dframe.rename(
        columns={
            f"STDPH{plane}": f"{ERR}{PHASE}{plane}",
            f"PH{plane}{MDL}": f"{PHASE}{plane}{MDL}",
            "S1": "S2",
        },
    )
    dframe[f"{DELTA}{PHASE}{plane}"] = df_ang_diff(dframe, f"{PHASE}{plane}", f"{PHASE}{plane}{MDL}")
    dframe[f"{ERR}{DELTA}{PHASE}{plane}"] = dframe.loc[:, f"{ERR}{PHASE}{plane}"].to_numpy()
    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_total_phase(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    suffix: str,
    plane: str,
    old_file_name: str = "phasetot",
    new_file_name: str = TOTAL_PHASE_NAME,
) -> None:
    """
    Looks in the provided directory for expected total phase file from ``BetaBeat.src`` for a given
    plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getphasetot(x,y).out, with the following expected columns: NAME, NAME2, S,
    S1, COUNT, PHASEX, STDPHX, PHXMDL, MUXMDL.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        suffix (str): Compensation suffix used in the provided BetaBeat.src output files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old total phase file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{suffix}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    dframe = dframe.rename(
        columns={
            f"STDPH{plane}": f"{ERR}{PHASE}{plane}",
            f"PH{plane}{MDL}": f"{PHASE}{plane}{MDL}",
            "S1": "S2",
        },
    )
    dframe[f"{DELTA}{PHASE}{plane}"] = df_ang_diff(dframe, f"{PHASE}{plane}", f"{PHASE}{plane}{MDL}")
    dframe[f"{ERR}{DELTA}{PHASE}{plane}"] = dframe.loc[:, f"{ERR}{PHASE}{plane}"].to_numpy()
    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_closed_orbit(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    plane: str,
    old_file_name: str = "CO",
    new_file_name: str = ORBIT_NAME,
) -> None:
    """
    Looks in the provided directory for expected closed orbit file from ``BetaBeat.src`` for a given
    plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getCO(x,y).out, with the following expected columns: NAME, S, COUNT,
    X, STDX, XMDL, MUXMDL.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old closed orbit file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    dframe = dframe.rename(columns={f"STD{plane}": f"{ERR}{plane}"})
    dframe[f"{DELTA}{plane}"] = df_diff(dframe, f"{plane}", f"{plane}{MDL}")
    dframe[f"{ERR}{DELTA}{plane}"] = dframe.loc[:, f"{ERR}{plane}"].to_numpy()
    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_dispersion(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    plane: str,
    old_file_name: str = "D",
    new_file_name: str = DISPERSION_NAME,
) -> None:
    """
    Looks in the provided directory for expected dispersion file from ``BetaBeat.src`` for a given
    plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getD(x,y).out, with the following expected columns: NAME, S, COUNT, DX,
    STDDX, DPX, DXMDL, DPXMDL, MUXMDL.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old dispersion file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    dframe = dframe.rename(columns={f"STDD{plane}": f"{ERR}D{plane}"})
    dframe[f"{DELTA}D{plane}"] = df_diff(dframe, f"D{plane}", f"D{plane}{MDL}")
    dframe[f"{ERR}{DELTA}D{plane}"] = dframe.loc[:, f"{ERR}D{plane}"].to_numpy()
    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_normalised_dispersion(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    plane: str,
    old_file_name: str = "ND",
    new_file_name: str = NORM_DISP_NAME,
) -> None:
    """
    Looks in the provided directory for expected normalized dispersion file from ``BetaBeat.src`` for a
    given plane, converts it to the output format used by ``omc3`` and  write them to the new location.

    The file naming should be getND(x,y).out, with the following expected columns: NAME, S, COUNT, NDX,
    STDNDX, DX, DPX, NDXMDL, DXMDL, DPXMDL, MUXMDL.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        plane (str): the transverse plane for which to look for the output file.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old normalized dispersion file")
    old_file_path = Path(inputdir) / f"get{old_file_name}{plane.lower()}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    dframe = dframe.rename(columns={f"STDND{plane}": f"{ERR}ND{plane}"})
    dframe[f"{DELTA}ND{plane}"] = df_diff(dframe, f"ND{plane}", f"ND{plane}{MDL}")
    dframe[f"{ERR}{DELTA}ND{plane}"] = dframe.loc[:, f"{ERR}ND{plane}"].to_numpy()

    if f"D{plane}" in dframe.columns:
        dframe = dframe.rename(columns={f"STDD{plane}": f"{ERR}D{plane}"})
        dframe[f"{DELTA}D{plane}"] = df_diff(dframe, f"D{plane}", f"D{plane}{MDL}")

    tfs.write(Path(outputdir) / f"{new_file_name}{plane.lower()}{EXT}", dframe)


def convert_old_coupling(
    inputdir: Union[Path, str],
    outputdir: Union[Path, str],
    suffix: str,
    old_file_name: str = "couple",
    new_file_name: str = "f"
) -> None:
    """
    Looks in the provided directory for expected coupling file from ``BetaBeat.src``, converts it to the
    output format used by ``omc3`` and  write them to the new location.

    The file naming should be getcouple(x,y).out, with the following expected columns: NAME, S, COUNT,
    F1001W, FWSTD1, F1001R, F1001I, F1010W, FWSTD2, F1010R, F1010I, Q1001, Q1001STD, Q1010, Q1010STD,
    MDLF1001R, MDLF1001I, MDLF1010R, MDLF1010I.

    Args:
        inputdir (Union[Path, str]): Location of the directory with BetaBeat.src output files.
        outputdir (Union[Path, str]): Location of the output directory for converted files.
        suffix (str): Compensation suffix used in the provided BetaBeat.src output files.
        old_file_name (str): the standard naming for the old output file.
        new_file_name (str): the standard naming for the new converted file.
    """
    LOGGER.info("Converting old coupling files")
    old_file_path = Path(inputdir) / f"get{old_file_name}{suffix}{OLD_EXT}"
    if not old_file_path.is_file():
        LOGGER.debug(f"Expected BetaBeat.src output at '{old_file_path.absolute()}' is not a file, skipping")
        return

    dframe = tfs.read(old_file_path)
    rdt_dfs = {
        "1001": dframe.loc[: , ["S", "COUNT", "F1001W", "FWSTD1", "F1001R", "F1001I",
                                "Q1001", "Q1001STD", "MDLF1001R", "MDLF1001I"]],
        "1010": dframe.loc[: , ["S", "COUNT", "F1010W", "FWSTD2", "F1010R", "F1010I",
                                "Q1010", "Q1010STD", "MDLF1010R", "MDLF1010I"]],
    }

    for i, rdt in enumerate(("1001", "1010")):
        LOGGER.debug(f"Converting F{rdt} file")
        rdt_dfs[rdt] = rdt_dfs[rdt].rename(
            columns={
                f"F{rdt}W": AMPLITUDE,
                f"FWSTD{i+1}": f"{ERR}{AMPLITUDE}",
                f"Q{rdt}": PHASE,
                f"Q{rdt}STD": f"{ERR}{PHASE}",
                f"F{rdt}R": REAL,
                f"F{rdt}I": IMAG,
                f"MDLF{rdt}R": f"{REAL}{MDL}",
                f"MDLF{rdt}I": f"{IMAG}{MDL}",
            }
        )
        tfs.write(Path(outputdir) / f"{new_file_name}{rdt}{EXT}", rdt_dfs[rdt])


if __name__ == "__main__":
    converter_entrypoint()
